# Copyright (C) 2008, 2009, 2011, 2013 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""DAV test server.

This defines the TestingDAVRequestHandler and the DAVServer classes which
implements the DAV specification parts used by the webdav plugin.
"""


import errno
import os
import re
import shutil  # FIXME: Can't we use breezy.osutils ?
import stat
import time
import urllib.parse  # FIXME: Can't we use breezy.urlutils ?

from breezy import trace, urlutils
from breezy.tests import http_server


class TestingDAVRequestHandler(http_server.TestingHTTPRequestHandler):
    """
    Subclass of TestingHTTPRequestHandler handling DAV requests.

    This is not a full implementation of a DAV server, only the parts
    really used by the plugin are.
    """

    _RANGE_HEADER_RE = re.compile(
        r'bytes (?P<begin>\d+)-(?P<end>\d+)/(?P<size>\d+|\*)')

    delete_success_code = 204
    move_default_overwrite = True


    def date_time_string(self, timestamp=None):
        """Return the current date and time formatted for a message header."""
        if timestamp is None:
            timestamp = time.time()
        year, month, day, hh, mm, ss, wd, y, z = time.gmtime(timestamp)
        s = "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (
                self.weekdayname[wd],
                day, self.monthname[month], year,
                hh, mm, ss)
        return s

    def _read(self, length):
        """Read the client socket"""
        return self.rfile.read(length)

    def _readline(self):
        """Read a full line on the client socket"""
        return self.rfile.readline()

    def read_body(self):
        """Read the body either by chunk or as a whole."""
        content_length = self.headers.get('Content-Length')
        encoding = self.headers.get('Transfer-Encoding')
        if encoding is not None:
            assert encoding == 'chunked'
            body = []
            # We receive the content by chunk
            while True:
                length, data = self.read_chunk()
                if length == 0:
                    break
                body.append(data)
            body = ''.join(body)

        else:
            if content_length is not None:
                body = self._read(int(content_length))

        return body

    def read_chunk(self):
        """Read a chunk of data.

        A chunk consists of:
        - a line containing the length of the data in hexa,
        - the data.
        - a empty line.

        An empty chunk specifies a length of zero
        """
        length = int(self._readline(), 16)
        data = None
        if length != 0:
            data = self._read(length)
            # Eats the newline following the chunk
            self._readline()
        return length, data

    def send_head(self):
        """Specialized version of SimpleHttpServer.

        We *don't* want the apache behavior of permanently redirecting
        directories without trailing slashes to directories with trailing
        slashes. That's a waste and a severe penalty for clients with high
        latency.

        The installation documentation of the plugin should mention the
        DirectorySlash apache directive and insists on turning it *Off*.
        """
        path = self.translate_path(self.path)
        f = None
        if os.path.isdir(path):
            for index in "index.html", "index.htm":
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            else:
                return self.list_directory(path)
        ctype = self.guess_type(path)
        if ctype.startswith('text/'):
            mode = 'r'
        else:
            mode = 'rb'
        try:
            f = open(path, mode)
        except OSError:
            self.send_error(404, "File not found")
            return None
        self.send_response(200)
        self.send_header("Content-type", ctype)
        fs = os.fstat(f.fileno())
        self.send_header("Content-Length", str(fs[6]))
        self.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
        self.end_headers()
        return f

    def do_PUT(self):
        """Serve a PUT request."""
        # FIXME: test_put_file_unicode makes us emit a traceback because a
        # UnicodeEncodeError occurs after the request headers have been sent
        # but before the body can be send. It's harmless and does not make the
        # test fails. Adressing that will mean protecting all reads from the
        # socket, which is too heavy for now -- vila 20070917
        path = self.translate_path(self.path)
        trace.mutter("do_PUT rel: [{}], abs: [{}]".format(self.path, path))

        do_append = False
        # Check the Content-Range header
        range_header = self.headers.get('Content-Range')
        if range_header is not None:
            match = self._RANGE_HEADER_RE.match(range_header)
            if match is None:
                # FIXME: RFC2616 says to return a 501 if we don't
                # understand the Content-Range header, but Apache
                # just ignores them (bad Apache).
                self.send_error(501, 'Not Implemented')
                return
            begin = int(match.group('begin'))
            do_append = True

        if self.headers.get('Expect') == '100-continue':
            # Tell the client to go ahead, we're ready to get the content
            self.send_response(100, "Continue")
            self.end_headers()

        try:
            trace.mutter("do_PUT will try to open: [%s]" % path)
            # Always write in binary mode.
            if do_append:
                f = open(path, 'ab')
                f.seek(begin)
            else:
                f = open(path, 'wb')
        except OSError as e:
            trace.mutter("do_PUT got: [%r] while opening/seeking on [%s]"
                         % (e, self.path))
            self.send_error(409, 'Conflict')
            return

        try:
            data = self.read_body()
            f.write(data)
        except OSError:
            # FIXME: We leave a partially written file here
            self.send_error(409, "Conflict")
            f.close()
            return
        f.close()
        trace.mutter("do_PUT done: [%s]" % self.path)
        self.send_response(201)
        self.end_headers()

    def do_MKCOL(self):
        """
        Serve a MKCOL request.

        MKCOL is an mkdir in DAV terminology for our part.
        """
        path = self.translate_path(self.path)
        trace.mutter("do_MKCOL rel: [{}], abs: [{}]".format(self.path, path))
        try:
            os.mkdir(path)
        except OSError as e:
            if e.errno in (errno.ENOENT, ):
                self.send_error(409, "Conflict")
            elif e.errno in (errno.EEXIST, errno.ENOTDIR):
                self.send_error(405, "Not allowed")
            else:
                # Ok we fail for an unnkown reason :-/
                raise
        else:
            self.send_response(201)
            self.end_headers()

    def do_COPY(self):
        """Serve a COPY request."""

        url_to = self.headers.get('Destination')
        if url_to is None:
            self.send_error(400, "Destination header missing")
            return
        (scheme, netloc, rel_to,
         params, query, fragment) = urllib.parse.urlparse(url_to)
        trace.mutter("urlparse: ({}) [{}]".format(url_to, rel_to))
        trace.mutter("do_COPY rel_from: [{}], rel_to: [{}]".format(self.path,
                                                               rel_to))
        abs_from = self.translate_path(self.path)
        abs_to = self.translate_path(rel_to)
        try:
            # TODO:  Check that rel_from  exists and  rel_to does
            # not.  In the  mean  time, just  go  along and  trap
            # exceptions
            shutil.copyfile(abs_from, abs_to)
        except OSError as e:
            if e.errno == errno.ENOENT:
                self.send_error(404, "File not found")
            else:
                self.send_error(409, "Conflict")
        else:
            # TODO: We may be able  to return 204 "No content" if
            # rel_to was existing (even  if the "No content" part
            # seems misleading, RFC2518 says so, stop arguing :)
            self.send_response(201)
            self.end_headers()

    def do_DELETE(self):
        """Serve a DELETE request.

        We don't implement a true DELETE as DAV defines it
        because we *should* fail to delete a non empty dir.
        """
        path = self.translate_path(self.path)
        trace.mutter("do_DELETE rel: [{}], abs: [{}]".format(self.path, path))
        try:
            # DAV  makes no  distinction between  files  and dirs
            # when required to nuke them,  but we have to. And we
            # also watch out for symlinks.
            real_path = os.path.realpath(path)
            if os.path.isdir(real_path):
                os.rmdir(path)
            else:
                os.remove(path)
        except OSError as e:
            if e.errno in (errno.ENOENT, ):
                self.send_error(404, "File not found")
            else:
                # Ok we fail for an unnkown reason :-/
                raise
        else:
            self.send_response(self.delete_success_code)
            self.end_headers()

    def do_MOVE(self):
        """Serve a MOVE request."""

        url_to = self.headers.get('Destination')
        if url_to is None:
            self.send_error(400, "Destination header missing")
            return
        overwrite_header = self.headers.get('Overwrite')
        should_overwrite = self.move_default_overwrite
        if overwrite_header == 'F':
            should_overwrite = False
        elif overwrite_header == 'T':
            should_overwrite = True
        (scheme, netloc, rel_to,
         params, query, fragment) = urllib.parse.urlparse(url_to)
        trace.mutter("urlparse: ({}) [{}]".format(url_to, rel_to))
        trace.mutter("do_MOVE rel_from: [{}], rel_to: [{}]".format(self.path,
                                                               rel_to))
        abs_from = self.translate_path(self.path)
        abs_to = self.translate_path(rel_to)
        if not should_overwrite and os.access(abs_to, os.F_OK):
            self.send_error(412, "Precondition Failed")
            return
        try:
            os.rename(abs_from, abs_to)
        except OSError as e:
            if e.errno == errno.ENOENT:
                self.send_error(404, "File not found")
            else:
                self.send_error(409, "Conflict")
        else:
            # TODO: We may be able  to return 204 "No content" if
            # rel_to was existing (even  if the "No content" part
            # seems misleading, RFC2518 says so, stop arguing :)
            self.send_response(201)
            self.end_headers()

    def _generate_response(self, path):
        local_path = self.translate_path(path)
        st = os.stat(local_path)
        prop = dict()

        def _prop(ns, name, value=None):
            if value is None:
                return '<{}:{}/>'.format(ns, name)
            else:
                return '<{}:{}>{}</{}:{}>'.format(ns, name, value, ns, name)

        # For namespaces (and test purposes), where apache2 use:
        # - lp1, we use liveprop,
        # - lp2, we use bzr
        if stat.S_ISDIR(st.st_mode):
            dpath = path
            if not dpath.endswith('/'):
                dpath += '/'
            prop['href'] = _prop('D', 'href', dpath)
            prop['type'] = _prop('liveprop', 'resourcetype', '<D:collection/>')
            prop['length'] = ''
            prop['exec'] = ''
        else:
            # FIXME: assert S_ISREG ? Handle symlinks ?
            prop['href'] = _prop('D', 'href', path)
            prop['type'] = _prop('liveprop', 'resourcetype')
            prop['length'] = _prop('liveprop', 'getcontentlength',
                                          st.st_size)
            if st.st_mode & stat.S_IXUSR:
                is_exec = 'T'
            else:
                is_exec = 'F'
            prop['exec'] = _prop('bzr', 'executable', is_exec)
        prop['status'] = _prop('D', 'status', 'HTTP/1.1 200 OK')

        response = """<D:response xmlns:liveprop="DAV:" xmlns:bzr="DAV:">
    %(href)s
    <D:propstat>
        <D:prop>
             %(type)s
             %(length)s
             %(exec)s
        </D:prop>
        %(status)s
    </D:propstat>
</D:response>
""" % prop
        return response, st

    def _generate_dir_responses(self, path, depth):
        local_path = self.translate_path(path)
        entries = os.listdir(local_path)

        for entry in entries:
            entry_path = urlutils.escape(entry)
            if path.endswith('/'):
                entry_path = path + entry_path
            else:
                entry_path = path + '/' + entry_path
            response, st = self._generate_response(entry_path)
            yield response
            if depth == 'Infinity' and stat.S_ISDIR(st.st_mode):
                yield from self._generate_dir_responses(entry_path, depth)

    def do_PROPFIND(self):
        """Serve a PROPFIND request."""
        depth = self.headers.get('Depth')
        if depth is None:
            depth = 'Infinity'
        if depth not in ('0', '1', 'Infinity'):
            self.send_error(400, "Bad Depth")
            return

        # Don't bother parsing the body, we handle only allprop anyway.
        # FIXME: Handle the body :)
        self.read_body()

        try:
            response, st = self._generate_response(self.path)
        except OSError as e:
            if e.errno == errno.ENOENT:
                self.send_error(404)
                return
            else:
                raise

        if depth in ('1', 'Infinity') and stat.S_ISDIR(st.st_mode):
            dir_responses = self._generate_dir_responses(self.path, depth)
        else:
            dir_responses = []

        # Generate the response, we don't care about performance, so we just
        # expand everything into a big string.
        response = ("""<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="DAV:">
{}{}
</D:multistatus>""".format(response, ''.join(dir_responses))).encode('utf-8')

        self.send_response(207)
        self.send_header('Content-length', len(response))
        self.end_headers()
        self.wfile.write(response)


class DAVServer(http_server.HttpServer):
    """Subclass of HttpServer that gives http+webdav urls.

    This is for use in testing: connections to this server will always go
    through _urllib where possible.
    """

    def __init__(self):
        # We    have   special    requests    to   handle    that
        # HttpServer_urllib doesn't know about
        super().__init__(TestingDAVRequestHandler)

    # urls returned by this server should require the webdav client impl
    _url_protocol = 'http+webdav'


class QuirkyTestingDAVRequestHandler(TestingDAVRequestHandler):
    """Various quirky/slightly off-spec behaviors.

    Used to test how gracefully we handle them.
    """

    delete_success_code = 200
    move_default_overwrite = False


class QuirkyDAVServer(http_server.HttpServer):
    """DAVServer implementing various quirky/slightly off-spec behaviors.

    Used to test how gracefully we handle them.
    """

    def __init__(self):
        # We    have   special    requests    to   handle    that
        # HttpServer_urllib doesn't know about
        super().__init__(QuirkyTestingDAVRequestHandler)

    # urls returned by this server should require the webdav client impl
    _url_protocol = 'http+webdav'




# Copyright (C) 2005-2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy.errors import CommandError, NoSuchRevision
from breezy.tests import TestCaseWithTransport
from breezy.workingtree import WorkingTree


class TestRevisionInfo(TestCaseWithTransport):

    def check_output(self, output, *args):
        """Verify that the expected output matches what brz says.

        The output is supplied first, so that you can supply a variable
        number of arguments to bzr.
        """
        self.assertEqual(self.run_bzr(*args)[0], output)

    def test_revision_info(self):
        """Test that 'brz revision-info' reports the correct thing."""
        wt = self.make_branch_and_tree('.')

        # Make history with a non-mainline rev
        wt.commit('Commit one', rev_id=b'a@r-0-1')
        wt.commit('Commit two', rev_id=b'a@r-0-1.1.1')
        wt.set_parent_ids([b'a@r-0-1', b'a@r-0-1.1.1'])
        wt.branch.set_last_revision_info(1, b'a@r-0-1')
        wt.commit('Commit three', rev_id=b'a@r-0-2')

        # This is expected to work even if the working tree is removed
        wt.controldir.destroy_workingtree()

        # Expected return values
        values = {
            '1': '1 a@r-0-1\n',
            '1.1.1': '1.1.1 a@r-0-1.1.1\n',
            '2': '2 a@r-0-2\n'
        }

        # Make sure with no arg it defaults to the head
        self.check_output(values['2'], 'revision-info')

        # Check the results of just specifying a numeric revision
        self.check_output(values['1'], 'revision-info 1')
        self.check_output(values['1.1.1'], 'revision-info 1.1.1')
        self.check_output(values['2'], 'revision-info 2')
        self.check_output(values['1'] + values['2'], 'revision-info 1 2')
        self.check_output('    ' + values['1']
                          + values['1.1.1']
                          + '    ' + values['2'],
                          'revision-info 1 1.1.1 2')
        self.check_output(values['2'] + values['1'], 'revision-info 2 1')

        # Check as above, only using the '--revision' syntax

        self.check_output(values['1'], 'revision-info -r 1')
        self.check_output(values['1.1.1'], 'revision-info --revision 1.1.1')
        self.check_output(values['2'], 'revision-info -r 2')
        self.check_output(values['1'] + values['2'], 'revision-info -r 1..2')
        self.check_output('    ' + values['1']
                          + values['1.1.1']
                          + '    ' + values['2'],
                          'revision-info -r 1..1.1.1..2')
        self.check_output(values['2'] + values['1'], 'revision-info -r 2..1')

        # Now try some more advanced revision specifications

        self.check_output(values['1'], 'revision-info -r revid:a@r-0-1')
        self.check_output(values['1.1.1'],
                          'revision-info --revision revid:a@r-0-1.1.1')

    def test_revision_info_explicit_branch_dir(self):
        """Test that 'brz revision-info' honors the '-d' option."""
        wt = self.make_branch_and_tree('branch')

        wt.commit('Commit one', rev_id=b'a@r-0-1')
        self.check_output('1 a@r-0-1\n', 'revision-info -d branch')

    def test_revision_info_tree(self):
        # Make branch and checkout
        wt = self.make_branch_and_tree('branch')
        wt.commit('Commit one', rev_id=b'a@r-0-1')

        # Make checkout and move the branch forward
        wt.branch.create_checkout('checkout', lightweight=True)
        wt.commit('Commit two', rev_id=b'a@r-0-2')

        # Make sure the checkout gives the right answer for branch and
        # tree
        self.check_output('2 a@r-0-2\n', 'revision-info -d checkout')
        self.check_output('1 a@r-0-1\n', 'revision-info --tree -d checkout')

    def test_revision_info_tree_no_working_tree(self):
        # Make branch with no tree
        b = self.make_branch('branch')

        # Try getting the --tree revision-info
        out, err = self.run_bzr('revision-info --tree -d branch', retcode=3)
        self.assertEqual('', out)
        self.assertEqual('brz: ERROR: No WorkingTree exists for "branch".\n',
                         err)

    def test_revision_info_not_in_history(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None))],
            revision_id=b'A-id')
        builder.build_snapshot([b'A-id'], [], revision_id=b'B-id')
        builder.build_snapshot([b'A-id'], [], revision_id=b'C-id')
        builder.finish_series()
        self.check_output('  1 A-id\n??? B-id\n  2 C-id\n',
                          'revision-info -d branch'
                          ' revid:A-id revid:B-id revid:C-id')



# Copyright (C) 2005-2008, 2010 by Canonical Ltd
#   Authors: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from unittest import TestLoader

from .... import config, tests
from ....bzr.bzrdir import BzrDir
from ....tests import TestCaseInTempDir
from ..emailer import EmailSender


def test_suite():
    return TestLoader().loadTestsFromName(__name__)


sample_config = (b"[DEFAULT]\n"
                 b"post_commit_to=demo@example.com\n"
                 b"post_commit_sender=Sample <foo@example.com>\n"
                 b"revision_mail_headers=X-Cheese: to the rescue!\n")

unconfigured_config = (b"[DEFAULT]\n"
                       b"email=Robert <foo@example.com>\n")

sender_configured_config = (b"[DEFAULT]\n"
                            b"post_commit_sender=Sample <foo@example.com>\n")

to_configured_config = (b"[DEFAULT]\n"
                        b"post_commit_to=Sample <foo@example.com>\n")

multiple_to_configured_config = (b"[DEFAULT]\n"
                                 b"post_commit_sender=Sender <from@example.com>\n"
                                 b"post_commit_to=Sample <foo@example.com>, Other <baz@bar.com>\n")

customized_mail_config = (b"[DEFAULT]\n"
                          b"post_commit_to=demo@example.com\n"
                          b"post_commit_sender=Sample <foo@example.com>\n"
                          b"post_commit_subject=[commit] $message\n"
                          b"post_commit_body='''$committer has committed "
                          b"revision 1 at $url.\n\n'''\n")

push_config = (b"[DEFAULT]\n"
               b"post_commit_to=demo@example.com\n"
               b"post_commit_push_pull=True\n")

with_url_config = (b"[DEFAULT]\n"
                   b"post_commit_url=http://some.fake/url/\n"
                   b"post_commit_to=demo@example.com\n"
                   b"post_commit_sender=Sample <foo@example.com>\n")

# FIXME: this should not use a literal log, rather grab one from breezy.log
sample_log = ('------------------------------------------------------------\n'
              'revno: 1\n'
              'revision-id: %s\n'
              'committer: Sample <john@example.com>\n'
              'branch nick: work\n'
              'timestamp: Thu 1970-01-01 00:00:01 +0000\n'
              'message:\n'
              '  foo bar baz\n'
              '  fuzzy\n'
              '  wuzzy\n')


class TestGetTo(TestCaseInTempDir):

    def test_body(self):
        sender, revid = self.get_sender()
        self.assertEqual(
            'At {}\n\n{}'.format(sender.url(), sample_log % revid.decode('utf-8')),
            sender.body())

    def test_custom_body(self):
        sender, revid = self.get_sender(customized_mail_config)
        self.assertEqual(
            '%s has committed revision 1 at %s.\n\n%s' %
            (sender.revision.committer, sender.url(), sample_log % revid.decode('utf-8')),
            sender.body())

    def test_command_line(self):
        sender, revid = self.get_sender()
        self.assertEqual(['mail', '-s', sender.subject(), '-a',
                          'From: ' + sender.from_address()] + sender.to(),
                         sender._command_line())

    def test_to(self):
        sender, revid = self.get_sender()
        self.assertEqual(['demo@example.com'], sender.to())

    def test_from(self):
        sender, revid = self.get_sender()
        self.assertEqual('Sample <foo@example.com>', sender.from_address())

    def test_from_default(self):
        sender, revid = self.get_sender(unconfigured_config)
        self.assertEqual('Robert <foo@example.com>', sender.from_address())

    def test_should_send(self):
        sender, revid = self.get_sender()
        self.assertEqual(True, sender.should_send())

    def test_should_not_send(self):
        sender, revid = self.get_sender(unconfigured_config)
        self.assertEqual(False, sender.should_send())

    def test_should_not_send_sender_configured(self):
        sender, revid = self.get_sender(sender_configured_config)
        self.assertEqual(False, sender.should_send())

    def test_should_not_send_to_configured(self):
        sender, revid = self.get_sender(to_configured_config)
        self.assertEqual(True, sender.should_send())

    def test_send_to_multiple(self):
        sender, revid = self.get_sender(multiple_to_configured_config)
        self.assertEqual(['Sample <foo@example.com>', 'Other <baz@bar.com>'],
                         sender.to())
        self.assertEqual(['Sample <foo@example.com>', 'Other <baz@bar.com>'],
                         sender._command_line()[-2:])

    def test_url_set(self):
        sender, revid = self.get_sender(with_url_config)
        self.assertEqual(sender.url(), 'http://some.fake/url/')

    def test_public_url_set(self):
        config = (b"[DEFAULT]\n"
                  b"public_branch=http://the.publication/location/\n")
        sender, revid = self.get_sender(config)
        self.assertEqual(sender.url(), 'http://the.publication/location/')

    def test_url_precedence(self):
        config = (b"[DEFAULT]\n"
                  b"post_commit_url=http://some.fake/url/\n"
                  b"public_branch=http://the.publication/location/\n")
        sender, revid = self.get_sender(config)
        self.assertEqual(sender.url(), 'http://some.fake/url/')

    def test_url_unset(self):
        sender, revid = self.get_sender()
        self.assertEqual(sender.url(), sender.branch.base)

    def test_subject(self):
        sender, revid = self.get_sender()
        self.assertEqual("Rev 1: foo bar baz in %s" %
                         sender.branch.base,
                         sender.subject())

    def test_custom_subject(self):
        sender, revid = self.get_sender(customized_mail_config)
        self.assertEqual("[commit] %s" %
                         sender.revision.get_summary(),
                         sender.subject())

    def test_diff_filename(self):
        sender, revid = self.get_sender()
        self.assertEqual('patch-1.diff', sender.diff_filename())

    def test_headers(self):
        sender, revid = self.get_sender()
        self.assertEqual({'X-Cheese': 'to the rescue!'},
                         sender.extra_headers())

    def get_sender(self, text=sample_config):
        my_config = config.MemoryStack(text)
        self.branch = BzrDir.create_branch_convenience('.')
        tree = self.branch.controldir.open_workingtree()
        revid = tree.commit('foo bar baz\nfuzzy\nwuzzy',
                    allow_pointless=True,
                    timestamp=1,
                    timezone=0,
                    committer="Sample <john@example.com>",
                    )
        sender = EmailSender(self.branch, revid, my_config)
        # This is usually only done after the EmailSender has locked the branch
        # and repository during send(), however, for testing, we need to do it
        # earlier, since send() is not called.
        sender._setup_revision_and_revno()
        return sender, revid


class TestEmailerWithLocal(tests.TestCaseWithTransport):
    """Test that Emailer will use a local branch if supplied."""

    def test_local_has_revision(self):
        master_tree = self.make_branch_and_tree('master')
        self.build_tree(['master/a'])
        master_tree.add('a')
        master_tree.commit('a')

        child_tree = master_tree.controldir.sprout('child').open_workingtree()
        child_tree.branch.bind(master_tree.branch)

        self.build_tree(['child/b'])
        child_tree.add(['b'])
        revision_id = child_tree.commit('b')

        sender = EmailSender(master_tree.branch, revision_id,
                             master_tree.branch.get_config(),
                             local_branch=child_tree.branch)

        # Make sure we are using the 'local_branch' repository, and not the
        # remote one.
        self.assertIs(child_tree.branch.repository, sender.repository)

    def test_local_missing_revision(self):
        master_tree = self.make_branch_and_tree('master')
        self.build_tree(['master/a'])
        master_tree.add('a')
        master_tree.commit('a')

        child_tree = master_tree.controldir.sprout('child').open_workingtree()
        child_tree.branch.bind(master_tree.branch)

        self.build_tree(['master/c'])
        master_tree.add(['c'])
        revision_id = master_tree.commit('c')

        self.assertFalse(
            child_tree.branch.repository.has_revision(revision_id))
        sender = EmailSender(master_tree.branch, revision_id,
                             master_tree.branch.get_config(),
                             local_branch=child_tree.branch)
        # We should be using the master repository here, because the child
        # repository doesn't contain the revision.
        self.assertIs(master_tree.branch.repository, sender.repository)



# Copyright (C) 2012-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Support for pristine tar deltas."""

import stat
from base64 import standard_b64decode

from dulwich.objects import Blob, Tree

README_CONTENTS = b"""\
This branch contains delta files that pristine-tar can use to
regenerate tarballs for its own releases.
"""


def revision_pristine_tar_data(rev):
    """Export the pristine tar data from a revision."""
    if 'deb-pristine-delta' in rev.properties:
        uuencoded = rev.properties['deb-pristine-delta']
        kind = 'gz'
    elif 'deb-pristine-delta-bz2' in rev.properties:
        uuencoded = rev.properties['deb-pristine-delta-bz2']
        kind = 'bz2'
    elif 'deb-pristine-delta-xz' in rev.properties:
        uuencoded = rev.properties['deb-pristine-delta-xz']
        kind = 'xz'
    else:
        raise KeyError(rev.revision_id)

    return (standard_b64decode(uuencoded), kind)


def get_pristine_tar_tree(repo):
    """Retrieve the pristine tar tree for a repository.

    """
    try:
        cid = repo.refs[b"refs/heads/pristine-tar"]
    except KeyError:
        return Tree()
    tid = repo.object_store[cid].tree
    return repo.object_store[tid]


def read_git_pristine_tar_data(repo, filename):
    """Read pristine data from a Git repository.

    :param repo: Git repository to read from
    :param filename: Name of file to read
    :return: Tuple with delta and id
    """
    tree = get_pristine_tar_tree(repo)
    delta = tree[filename + b".delta"][1]
    gitid = tree[filename + b".id"][1]
    return (repo.object_store[delta].data,
            repo.object_store[gitid].data)


def store_git_pristine_tar_data(repo, filename, delta, gitid,
                                message=None, **kwargs):
    """Add pristine tar data to a Git repository.

    :param repo: Git repository to add data to
    :param filename: Name of file to store for
    :param delta: pristine-tar delta
    :param gitid: Git id the pristine tar delta is generated against
    """
    delta_ob = Blob.from_string(delta)
    delta_name = filename + b".delta"
    id_ob = Blob.from_string(gitid)
    id_name = filename + b".id"
    objects = [
        (delta_ob, delta_name),
        (id_ob, id_name)]
    tree = get_pristine_tar_tree(repo)
    tree.add(delta_name, stat.S_IFREG | 0o644, delta_ob.id)
    tree.add(id_name, stat.S_IFREG | 0o644, id_ob.id)
    if b"README" not in tree:
        readme_ob = Blob.from_string(README_CONTENTS)
        objects.append((readme_ob, b"README"))
        tree.add(b"README", stat.S_IFREG | 0o644, readme_ob.id)
    objects.append((tree, ""))
    repo.object_store.add_objects(objects)
    if message is None:
        message = b'pristine-tar data for %s' % filename
    return repo.do_commit(ref=b'refs/heads/pristine-tar', tree=tree.id,
                          message=message, **kwargs)



#    __init__.py -- Quilt support for breezy
#
#    Breezy is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Breezy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Breezy; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Quilt patch integration.

This plugin adds support for three configuration options:

 * quilt.commit_policy
 * quilt.smart_merge
 * quilt.tree_policy

"""

from ... import trace
from ...errors import BzrError


class QuiltUnapplyError(BzrError):

    _fmt = ("Unable to unapply quilt patches for %(kind)r tree: %(msg)s")

    def __init__(self, kind, msg):
        BzrError.__init__(self)
        self.kind = kind
        if msg is not None and msg.count("\n") == 1:
            msg = msg.strip()
        self.msg = msg


def pre_merge_quilt(merger):
    if getattr(merger, "_no_quilt_unapplying", False):
        return

    config = merger.working_tree.get_config_stack()
    merger.debuild_config = config
    if not config.get('quilt.smart_merge'):
        trace.mutter("skipping smart quilt merge, not enabled.")
        return

    if (not merger.other_tree.is_versioned(".pc") and
            not merger.this_tree.is_versioned(".pc") and
            not merger.working_tree.is_versioned(".pc")):
        return

    from .quilt import QuiltError, QuiltPatches
    quilt = QuiltPatches(merger.working_tree)
    from .merge import tree_unapply_patches
    trace.note("Unapplying quilt patches to prevent spurious conflicts")
    merger._quilt_tempdirs = []
    merger._old_quilt_series = quilt.series()
    if merger._old_quilt_series:
        quilt.pop_all()
    try:
        merger.this_tree, this_dir = tree_unapply_patches(
            merger.this_tree, merger.this_branch, force=True)
    except QuiltError as e:
        raise QuiltUnapplyError("this", e.stderr)
    else:
        if this_dir is not None:
            merger._quilt_tempdirs.append(this_dir)
    try:
        merger.base_tree, base_dir = tree_unapply_patches(
            merger.base_tree, merger.this_branch, force=True)
    except QuiltError as e:
        raise QuiltUnapplyError("base", e.stderr)
    else:
        if base_dir is not None:
            merger._quilt_tempdirs.append(base_dir)
    other_branch = getattr(merger, "other_branch", None)
    if other_branch is None:
        other_branch = merger.this_branch
    try:
        merger.other_tree, other_dir = tree_unapply_patches(
            merger.other_tree, other_branch, force=True)
    except QuiltError as e:
        raise QuiltUnapplyError("other", e.stderr)
    else:
        if other_dir is not None:
            merger._quilt_tempdirs.append(other_dir)


def post_merge_quilt_cleanup(merger):
    import shutil
    for dir in getattr(merger, "_quilt_tempdirs", []):
        shutil.rmtree(dir)
    config = merger.working_tree.get_config_stack()
    policy = config.get('quilt.tree_policy')
    if policy is None:
        return
    from .merge import post_process_quilt_patches
    post_process_quilt_patches(
        merger.working_tree,
        getattr(merger, "_old_quilt_series", []), policy)


def start_commit_check_quilt(tree):
    """start_commit hook which checks the state of quilt patches.
    """
    config = tree.get_config_stack()
    policy = config.get('quilt.commit_policy')
    from .merge import start_commit_quilt_patches
    from .wrapper import QuiltNotInstalled
    try:
        start_commit_quilt_patches(tree, policy)
    except QuiltNotInstalled:
        trace.warning(
            'quilt not installed; not updating patches')


def post_build_tree_quilt(tree):
    config = tree.get_config_stack()
    policy = config.get('quilt.tree_policy')
    if policy is None:
        return
    from .merge import post_process_quilt_patches
    from .wrapper import QuiltNotInstalled
    try:
        post_process_quilt_patches(tree, [], policy)
    except QuiltNotInstalled:
        trace.warning(
            'quilt not installed; not touching patches')


from ...hooks import install_lazy_named_hook

install_lazy_named_hook(
    "breezy.merge", "Merger.hooks",
    'pre_merge_quilt', pre_merge_quilt,
    'Quilt patch (un)applying')
install_lazy_named_hook(
    "breezy.mutabletree", "MutableTree.hooks",
    "start_commit", start_commit_check_quilt,
    "Check for (un)applied quilt patches")
install_lazy_named_hook(
    "breezy.merge", "Merger.hooks",
    'post_merge', post_merge_quilt_cleanup,
    'Cleaning up quilt temporary directories')
install_lazy_named_hook(
    "breezy.mutabletree", "MutableTree.hooks",
    'post_build_tree', post_build_tree_quilt,
    'Applying quilt patches.')


from ...config import Option, bool_from_store, option_registry

option_registry.register(
    Option('quilt.smart_merge', default=True, from_unicode=bool_from_store,
           help="Unapply quilt patches before merging."))


def policy_from_store(s):
    if s not in ('applied', 'unapplied'):
        raise ValueError('Invalid quilt.commit_policy: %s' % s)
    return s

option_registry.register(
    Option('quilt.commit_policy', default=None, from_unicode=policy_from_store,
           help="Whether to apply or unapply all patches in commits."))

option_registry.register(
    Option('quilt.tree_policy', default=None, from_unicode=policy_from_store,
           help="Whether to apply or unapply all patches after checkout/update."))


def load_tests(loader, basic_tests, pattern):
    basic_tests.addTest(
        loader.loadTestsFromModuleNames([__name__ + '.tests']))
    return basic_tests



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Server-side bzrdir related request implmentations."""

import fastbencode as bencode

from ... import branch, errors, repository, urlutils
from ...controldir import network_format_registry
from .. import BzrProber
from ..bzrdir import BzrDir, BzrDirFormat
from .request import (FailedSmartServerResponse, SmartServerRequest,
                      SuccessfulSmartServerResponse)


class SmartServerRequestOpenBzrDir(SmartServerRequest):

    def do(self, path):
        try:
            t = self.transport_from_client_path(path)
        except errors.PathNotChild:
            # The client is trying to ask about a path that they have no access
            # to.
            # Ideally we'd return a FailedSmartServerResponse here rather than
            # a "successful" negative, but we want to be compatibile with
            # clients that don't anticipate errors from this method.
            answer = b'no'
        else:
            bzr_prober = BzrProber()
            try:
                bzr_prober.probe_transport(t)
            except (errors.NotBranchError, errors.UnknownFormatError):
                answer = b'no'
            else:
                answer = b'yes'
        return SuccessfulSmartServerResponse((answer,))


class SmartServerRequestOpenBzrDir_2_1(SmartServerRequest):

    def do(self, path):
        """Is there a BzrDir present, and if so does it have a working tree?

        New in 2.1.
        """
        try:
            t = self.transport_from_client_path(path)
        except errors.PathNotChild:
            # The client is trying to ask about a path that they have no access
            # to.
            return SuccessfulSmartServerResponse((b'no',))
        try:
            bd = BzrDir.open_from_transport(t)
        except errors.NotBranchError:
            answer = (b'no',)
        else:
            answer = (b'yes',)
            if bd.has_workingtree():
                answer += (b'yes',)
            else:
                answer += (b'no',)
        return SuccessfulSmartServerResponse(answer)


class SmartServerRequestBzrDir(SmartServerRequest):

    def do(self, path, *args):
        """Open a BzrDir at path, and return `self.do_bzrdir_request(*args)`."""
        try:
            self._bzrdir = BzrDir.open_from_transport(
                self.transport_from_client_path(path))
        except errors.NotBranchError as e:
            return FailedSmartServerResponse((b'nobranch',))
        return self.do_bzrdir_request(*args)

    def _boolean_to_yes_no(self, a_boolean):
        if a_boolean:
            return b'yes'
        else:
            return b'no'

    def _format_to_capabilities(self, repo_format):
        rich_root = self._boolean_to_yes_no(repo_format.rich_root_data)
        tree_ref = self._boolean_to_yes_no(
            repo_format.supports_tree_reference)
        external_lookup = self._boolean_to_yes_no(
            repo_format.supports_external_lookups)
        return rich_root, tree_ref, external_lookup

    def _repo_relpath(self, current_transport, repository):
        """Get the relative path for repository from current_transport."""
        # the relpath of the bzrdir in the found repository gives us the
        # path segments to pop-out.
        relpath = repository.user_transport.relpath(
            current_transport.base)
        if len(relpath):
            segments = ['..'] * len(relpath.split('/'))
        else:
            segments = []
        return '/'.join(segments)


class SmartServerBzrDirRequestDestroyBranch(SmartServerRequestBzrDir):

    def do_bzrdir_request(self, name=None):
        """Destroy the branch with the specified name.

        New in 2.5.0.
        :return: On success, 'ok'.
        """
        try:
            self._bzrdir.destroy_branch(
                name.decode('utf-8') if name is not None else None)
        except errors.NotBranchError as e:
            return FailedSmartServerResponse((b'nobranch',))
        return SuccessfulSmartServerResponse((b'ok',))


class SmartServerBzrDirRequestHasWorkingTree(SmartServerRequestBzrDir):

    def do_bzrdir_request(self, name=None):
        """Check whether there is a working tree present.

        New in 2.5.0.

        :return: If there is a working tree present, 'yes'.
            Otherwise 'no'.
        """
        if self._bzrdir.has_workingtree():
            return SuccessfulSmartServerResponse((b'yes', ))
        else:
            return SuccessfulSmartServerResponse((b'no', ))


class SmartServerBzrDirRequestDestroyRepository(SmartServerRequestBzrDir):

    def do_bzrdir_request(self, name=None):
        """Destroy the repository.

        New in 2.5.0.

        :return: On success, 'ok'.
        """
        try:
            self._bzrdir.destroy_repository()
        except errors.NoRepositoryPresent as e:
            return FailedSmartServerResponse((b'norepository',))
        return SuccessfulSmartServerResponse((b'ok',))


class SmartServerBzrDirRequestCloningMetaDir(SmartServerRequestBzrDir):

    def do_bzrdir_request(self, require_stacking):
        """Get the format that should be used when cloning from this dir.

        New in 1.13.

        :return: on success, a 3-tuple of network names for (control,
            repository, branch) directories, where '' signifies "not present".
            If this BzrDir contains a branch reference then this will fail with
            BranchReference; clients should resolve branch references before
            calling this RPC.
        """
        try:
            branch_ref = self._bzrdir.get_branch_reference()
        except errors.NotBranchError:
            branch_ref = None
        if branch_ref is not None:
            # The server shouldn't try to resolve references, and it quite
            # possibly can't reach them anyway.  The client needs to resolve
            # the branch reference to determine the cloning_metadir.
            return FailedSmartServerResponse((b'BranchReference',))
        if require_stacking == b"True":
            require_stacking = True
        else:
            require_stacking = False
        control_format = self._bzrdir.cloning_metadir(
            require_stacking=require_stacking)
        control_name = control_format.network_name()
        if not control_format.fixed_components:
            branch_name = (b'branch',
                           control_format.get_branch_format().network_name())
            repository_name = control_format.repository_format.network_name()
        else:
            # Only MetaDir has delegated formats today.
            branch_name = (b'branch', b'')
            repository_name = b''
        return SuccessfulSmartServerResponse((control_name, repository_name,
                                              branch_name))


class SmartServerBzrDirRequestCheckoutMetaDir(SmartServerRequestBzrDir):
    """Get the format to use for checkouts.

    New in 2.5.

    :return: on success, a 3-tuple of network names for (control,
        repository, branch) directories, where '' signifies "not present".
        If this BzrDir contains a branch reference then this will fail with
        BranchReference; clients should resolve branch references before
        calling this RPC (they should not try to create a checkout of a
        checkout).
    """

    def do_bzrdir_request(self):
        try:
            branch_ref = self._bzrdir.get_branch_reference()
        except errors.NotBranchError:
            branch_ref = None
        if branch_ref is not None:
            # The server shouldn't try to resolve references, and it quite
            # possibly can't reach them anyway.  The client needs to resolve
            # the branch reference to determine the cloning_metadir.
            return FailedSmartServerResponse((b'BranchReference',))
        control_format = self._bzrdir.checkout_metadir()
        control_name = control_format.network_name()
        if not control_format.fixed_components:
            branch_name = control_format.get_branch_format().network_name()
            repo_name = control_format.repository_format.network_name()
        else:
            branch_name = b''
            repo_name = b''
        return SuccessfulSmartServerResponse(
            (control_name, repo_name, branch_name))


class SmartServerRequestCreateBranch(SmartServerRequestBzrDir):

    def do(self, path, network_name):
        """Create a branch in the bzr dir at path.

        This operates precisely like 'bzrdir.create_branch'.

        If a bzrdir is not present, an exception is propogated
        rather than 'no branch' because these are different conditions (and
        this method should only be called after establishing that a bzr dir
        exists anyway).

        This is the initial version of this method introduced to the smart
        server for 1.13.

        :param path: The path to the bzrdir.
        :param network_name: The network name of the branch type to create.
        :return: ('ok', branch_format, repo_path, rich_root, tree_ref,
            external_lookup, repo_format)
        """
        bzrdir = BzrDir.open_from_transport(
            self.transport_from_client_path(path))
        format = branch.network_format_registry.get(network_name)
        bzrdir.branch_format = format
        result = format.initialize(bzrdir, name="")
        rich_root, tree_ref, external_lookup = self._format_to_capabilities(
            result.repository._format)
        branch_format = result._format.network_name()
        repo_format = result.repository._format.network_name()
        repo_path = self._repo_relpath(bzrdir.root_transport,
                                       result.repository)
        # branch format, repo relpath, rich_root, tree_ref, external_lookup,
        # repo_network_name
        return SuccessfulSmartServerResponse((b'ok', branch_format, repo_path,
                                              rich_root, tree_ref, external_lookup, repo_format))


class SmartServerRequestCreateRepository(SmartServerRequestBzrDir):

    def do(self, path, network_name, shared):
        """Create a repository in the bzr dir at path.

        This operates precisely like 'bzrdir.create_repository'.

        If a bzrdir is not present, an exception is propagated
        rather than 'no branch' because these are different conditions (and
        this method should only be called after establishing that a bzr dir
        exists anyway).

        This is the initial version of this method introduced to the smart
        server for 1.13.

        :param path: The path to the bzrdir.
        :param network_name: The network name of the repository type to create.
        :param shared: The value to pass create_repository for the shared
            parameter.
        :return: (ok, rich_root, tree_ref, external_lookup, network_name)
        """
        bzrdir = BzrDir.open_from_transport(
            self.transport_from_client_path(path))
        shared = shared == b'True'
        format = repository.network_format_registry.get(network_name)
        bzrdir.repository_format = format
        result = format.initialize(bzrdir, shared=shared)
        rich_root, tree_ref, external_lookup = self._format_to_capabilities(
            result._format)
        return SuccessfulSmartServerResponse((b'ok', rich_root, tree_ref,
                                              external_lookup, result._format.network_name()))


class SmartServerRequestFindRepository(SmartServerRequestBzrDir):

    def _find(self, path):
        """try to find a repository from path upwards

        This operates precisely like 'bzrdir.find_repository'.

        :return: (relpath, rich_root, tree_ref, external_lookup, network_name).
            All are strings, relpath is a / prefixed path, the next three are
            either 'yes' or 'no', and the last is a repository format network
            name.
        :raises errors.NoRepositoryPresent: When there is no repository
            present.
        """
        bzrdir = BzrDir.open_from_transport(
            self.transport_from_client_path(path))
        repository = bzrdir.find_repository()
        path = self._repo_relpath(bzrdir.root_transport, repository)
        rich_root, tree_ref, external_lookup = self._format_to_capabilities(
            repository._format)
        network_name = repository._format.network_name()
        return path, rich_root, tree_ref, external_lookup, network_name


class SmartServerRequestFindRepositoryV1(SmartServerRequestFindRepository):

    def do(self, path):
        """try to find a repository from path upwards

        This operates precisely like 'bzrdir.find_repository'.

        If a bzrdir is not present, an exception is propagated
        rather than 'no branch' because these are different conditions.

        This is the initial version of this method introduced with the smart
        server. Modern clients will try the V2 method that adds support for the
        supports_external_lookups attribute.

        :return: norepository or ok, relpath.
        """
        try:
            path, rich_root, tree_ref, external_lookup, name = self._find(path)
            return SuccessfulSmartServerResponse((b'ok', path.encode('utf-8'), rich_root, tree_ref))
        except errors.NoRepositoryPresent:
            return FailedSmartServerResponse((b'norepository', ))


class SmartServerRequestFindRepositoryV2(SmartServerRequestFindRepository):

    def do(self, path):
        """try to find a repository from path upwards

        This operates precisely like 'bzrdir.find_repository'.

        If a bzrdir is not present, an exception is propagated
        rather than 'no branch' because these are different conditions.

        This is the second edition of this method introduced in bzr 1.3, which
        returns information about the supports_external_lookups format
        attribute too.

        :return: norepository or ok, relpath, rich_root, tree_ref,
            external_lookup.
        """
        try:
            path, rich_root, tree_ref, external_lookup, name = self._find(path)
            return SuccessfulSmartServerResponse(
                (b'ok', path.encode('utf-8'), rich_root, tree_ref, external_lookup))
        except errors.NoRepositoryPresent:
            return FailedSmartServerResponse((b'norepository', ))


class SmartServerRequestFindRepositoryV3(SmartServerRequestFindRepository):

    def do(self, path):
        """try to find a repository from path upwards

        This operates precisely like 'bzrdir.find_repository'.

        If a bzrdir is not present, an exception is propogated
        rather than 'no branch' because these are different conditions.

        This is the third edition of this method introduced in bzr 1.13, which
        returns information about the network name of the repository format.

        :return: norepository or ok, relpath, rich_root, tree_ref,
            external_lookup, network_name.
        """
        try:
            path, rich_root, tree_ref, external_lookup, name = self._find(path)
            return SuccessfulSmartServerResponse(
                (b'ok', path.encode('utf-8'), rich_root, tree_ref, external_lookup, name))
        except errors.NoRepositoryPresent:
            return FailedSmartServerResponse((b'norepository', ))


class SmartServerBzrDirRequestConfigFile(SmartServerRequestBzrDir):

    def do_bzrdir_request(self):
        """Get the configuration bytes for a config file in bzrdir.

        The body is not utf8 decoded - it is the literal bytestream from disk.
        """
        config = self._bzrdir._get_config()
        if config is None:
            content = b''
        else:
            content = config._get_config_file().read()
        return SuccessfulSmartServerResponse((), content)


class SmartServerBzrDirRequestGetBranches(SmartServerRequestBzrDir):

    def do_bzrdir_request(self):
        """Get the branches in a control directory.

        The body is a bencoded dictionary, with values similar to the return
        value of the open branch request.
        """
        branch_names = self._bzrdir.branch_names()
        ret = {}
        for name in branch_names:
            if name is None:
                name = b""
            branch_ref = self._bzrdir.get_branch_reference(name=name)
            if branch_ref is not None:
                branch_ref = urlutils.relative_url(self._bzrdir.user_url, branch_ref)
                value = (b"ref", branch_ref.encode('utf-8'))
            else:
                b = self._bzrdir.open_branch(name=name, ignore_fallbacks=True)
                value = (b"branch", b._format.network_name())
            ret[name.encode('utf-8')] = value
        return SuccessfulSmartServerResponse(
            (b"success", ), bencode.bencode(ret))


class SmartServerRequestInitializeBzrDir(SmartServerRequest):

    def do(self, path):
        """Initialize a bzrdir at path.

        The default format of the server is used.
        :return: SmartServerResponse(('ok', ))
        """
        target_transport = self.transport_from_client_path(path)
        BzrDirFormat.get_default_format().initialize_on_transport(target_transport)
        return SuccessfulSmartServerResponse((b'ok', ))


class SmartServerRequestBzrDirInitializeEx(SmartServerRequestBzrDir):

    def parse_NoneTrueFalse(self, arg):
        if not arg:
            return None
        if arg == b'False':
            return False
        if arg == b'True':
            return True
        raise AssertionError("invalid arg %r" % arg)

    def parse_NoneBytestring(self, arg):
        return arg or None

    def parse_NoneString(self, arg):
        if not arg:
            return None
        return arg.decode('utf-8')

    def _serialize_NoneTrueFalse(self, arg):
        if arg is False:
            return b'False'
        if not arg:
            return b''
        return b'True'

    def do(self, bzrdir_network_name, path, use_existing_dir, create_prefix,
           force_new_repo, stacked_on, stack_on_pwd, repo_format_name,
           make_working_trees, shared_repo):
        """Initialize a bzrdir at path as per
        BzrDirFormat.initialize_on_transport_ex.

        New in 1.16.  (Replaces BzrDirFormat.initialize_ex verb from 1.15).

        :return: return SuccessfulSmartServerResponse((repo_path, rich_root,
            tree_ref, external_lookup, repo_network_name,
            repo_bzrdir_network_name, bzrdir_format_network_name,
            NoneTrueFalse(stacking), final_stack, final_stack_pwd,
            repo_lock_token))
        """
        target_transport = self.transport_from_client_path(path)
        format = network_format_registry.get(bzrdir_network_name)
        use_existing_dir = self.parse_NoneTrueFalse(use_existing_dir)
        create_prefix = self.parse_NoneTrueFalse(create_prefix)
        force_new_repo = self.parse_NoneTrueFalse(force_new_repo)
        stacked_on = self.parse_NoneString(stacked_on)
        stack_on_pwd = self.parse_NoneString(stack_on_pwd)
        make_working_trees = self.parse_NoneTrueFalse(make_working_trees)
        shared_repo = self.parse_NoneTrueFalse(shared_repo)
        if stack_on_pwd == b'.':
            stack_on_pwd = target_transport.base.encode('utf-8')
        repo_format_name = self.parse_NoneBytestring(repo_format_name)
        repo, bzrdir, stacking, repository_policy = \
            format.initialize_on_transport_ex(target_transport,
                                              use_existing_dir=use_existing_dir, create_prefix=create_prefix,
                                              force_new_repo=force_new_repo, stacked_on=stacked_on,
                                              stack_on_pwd=stack_on_pwd, repo_format_name=repo_format_name,
                                              make_working_trees=make_working_trees, shared_repo=shared_repo)
        if repo is None:
            repo_path = ''
            repo_name = b''
            rich_root = tree_ref = external_lookup = b''
            repo_bzrdir_name = b''
            final_stack = None
            final_stack_pwd = None
            repo_lock_token = b''
        else:
            repo_path = self._repo_relpath(bzrdir.root_transport, repo)
            if repo_path == '':
                repo_path = '.'
            rich_root, tree_ref, external_lookup = self._format_to_capabilities(
                repo._format)
            repo_name = repo._format.network_name()
            repo_bzrdir_name = repo.controldir._format.network_name()
            final_stack = repository_policy._stack_on
            final_stack_pwd = repository_policy._stack_on_pwd
            # It is returned locked, but we need to do the lock to get the lock
            # token.
            repo.unlock()
            repo_lock_token = repo.lock_write().repository_token or b''
            if repo_lock_token:
                repo.leave_lock_in_place()
            repo.unlock()
        final_stack = final_stack or ''
        final_stack_pwd = final_stack_pwd or ''

        # We want this to be relative to the bzrdir.
        if final_stack_pwd:
            final_stack_pwd = urlutils.relative_url(
                target_transport.base, final_stack_pwd)

        # Can't meaningfully return a root path.
        if final_stack.startswith('/'):
            client_path = self._root_client_path + final_stack[1:]
            final_stack = urlutils.relative_url(
                self._root_client_path, client_path)
            final_stack_pwd = '.'

        return SuccessfulSmartServerResponse((repo_path.encode('utf-8'),
                                              rich_root, tree_ref, external_lookup, repo_name, repo_bzrdir_name,
                                              bzrdir._format.network_name(),
                                              self._serialize_NoneTrueFalse(
                                                  stacking), final_stack.encode('utf-8'),
                                              final_stack_pwd.encode('utf-8'), repo_lock_token))


class SmartServerRequestOpenBranch(SmartServerRequestBzrDir):

    def do_bzrdir_request(self):
        """open a branch at path and return the branch reference or branch."""
        try:
            reference_url = self._bzrdir.get_branch_reference()
            if reference_url is None:
                reference_url = ''
            return SuccessfulSmartServerResponse((b'ok', reference_url.encode('utf-8')))
        except errors.NotBranchError as e:
            return FailedSmartServerResponse((b'nobranch',))


class SmartServerRequestOpenBranchV2(SmartServerRequestBzrDir):

    def do_bzrdir_request(self):
        """open a branch at path and return the reference or format."""
        try:
            reference_url = self._bzrdir.get_branch_reference()
            if reference_url is None:
                br = self._bzrdir.open_branch(ignore_fallbacks=True)
                format = br._format.network_name()
                return SuccessfulSmartServerResponse((b'branch', format))
            else:
                return SuccessfulSmartServerResponse((b'ref', reference_url.encode('utf-8')))
        except errors.NotBranchError as e:
            return FailedSmartServerResponse((b'nobranch',))


class SmartServerRequestOpenBranchV3(SmartServerRequestBzrDir):

    def do_bzrdir_request(self):
        """Open a branch at path and return the reference or format.

        This version introduced in 2.1.

        Differences to SmartServerRequestOpenBranchV2:
          * can return 2-element ('nobranch', extra), where 'extra' is a string
            with an explanation like 'location is a repository'.  Previously
            a 'nobranch' response would never have more than one element.
        """
        try:
            reference_url = self._bzrdir.get_branch_reference()
            if reference_url is None:
                br = self._bzrdir.open_branch(ignore_fallbacks=True)
                format = br._format.network_name()
                return SuccessfulSmartServerResponse((b'branch', format))
            else:
                return SuccessfulSmartServerResponse((b'ref', reference_url.encode('utf-8')))
        except errors.NotBranchError as e:
            # Stringify the exception so that its .detail attribute will be
            # filled out.
            str(e)
            resp = (b'nobranch',)
            detail = e.detail
            if detail:
                if detail.startswith(': '):
                    detail = detail[2:]
                resp += (detail.encode('utf-8'),)
            return FailedSmartServerResponse(resp)



# Copyright (C) 2006, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os

from breezy import osutils, tests, workingtree


class TestJoin(tests.TestCaseWithTransport):

    def make_trees(self):
        base_tree = self.make_branch_and_tree('tree',
                                              format='development-subtree')
        base_tree.commit('empty commit')
        self.build_tree(['tree/subtree/', 'tree/subtree/file1'])
        sub_tree = self.make_branch_and_tree('tree/subtree')
        sub_tree.add('file1', ids=b'file1-id')
        sub_tree.commit('added file1')
        return base_tree, sub_tree

    def check_success(self, path):
        base_tree = workingtree.WorkingTree.open(path)
        self.assertEqual(b'file1-id', base_tree.path2id('subtree/file1'))

    def test_join(self):
        base_tree, sub_tree = self.make_trees()
        self.run_bzr('join tree/subtree')
        self.check_success('tree')

    def test_join_dot(self):
        base_tree, sub_tree = self.make_trees()
        self.run_bzr('join .', working_dir='tree/subtree')
        self.check_success('tree')

    def test_join_error(self):
        base_tree, sub_tree = self.make_trees()
        os.mkdir('tree/subtree2')
        osutils.rename('tree/subtree', 'tree/subtree2/subtree')
        self.run_bzr_error(
            ('Cannot join .*subtree.  Parent directory is not versioned',),
            'join tree/subtree2/subtree')
        # disabled because this gives an ugly error at present -- mbp 20070306
        # self.run_bzr_error(
        ##     ('Cannot join .*subtree.  Parent directory is not versioned',),
        # 'join', '--reference', 'tree/subtree2/subtree')
        self.run_bzr_error(('Not a branch:.*subtree2',),
                           'join tree/subtree2')

    def test_join_reference(self):
        """Join can add a reference if --reference is supplied"""
        base_tree, sub_tree = self.make_trees()
        subtree_root_id = sub_tree.path2id('')
        self.run_bzr('join . --reference', working_dir='tree/subtree')
        sub_tree.lock_read()
        self.addCleanup(sub_tree.unlock)
        if sub_tree.supports_setting_file_ids():
            self.assertEqual(b'file1-id', sub_tree.path2id('file1'))
            self.assertEqual('file1', sub_tree.id2path(b'file1-id'))
            self.assertEqual(subtree_root_id, sub_tree.path2id(''))
            self.assertEqual('', sub_tree.id2path(subtree_root_id))
            self.assertEqual(
                sub_tree.path2id('file1'), base_tree.path2id('subtree/file1'))

        base_tree.lock_read()
        self.addCleanup(base_tree.unlock)
        self.assertEqual(['subtree'], list(base_tree.iter_references()))
        if base_tree.supports_setting_file_ids():
            self.assertEqual(b'file1-id', sub_tree.path2id('file1'))
            self.assertEqual('file1', sub_tree.id2path(b'file1-id'))
            self.assertEqual(subtree_root_id, base_tree.path2id('subtree'))
            self.assertEqual('subtree', base_tree.id2path(subtree_root_id))

    def test_references_check_repository_support(self):
        """Users are stopped from adding a reference that can't be committed."""
        # in 0.15 the default format has a dirstate workingtree, that can
        # support tree references, but the default repository format
        # cannot.
        tree = self.make_branch_and_tree('tree', format='dirstate')
        tree2 = self.make_branch_and_tree('tree/subtree')
        out, err = self.run_bzr('join --reference tree/subtree',
                                retcode=3)
        self.assertContainsRe(err, r"Can't join trees")
        self.assertContainsRe(err, r"use brz upgrade")



# Copyright (C) 2005, 2006, 2007, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys


def shellcomplete(context=None, outfile=None):
    if outfile is None:
        outfile = sys.stdout
    if context is None:
        shellcomplete_commands(outfile=outfile)
    else:
        shellcomplete_on_command(context, outfile=outfile)


def shellcomplete_on_command(cmdname, outfile=None):
    cmdname = str(cmdname)

    if outfile is None:
        outfile = sys.stdout

    from inspect import getdoc

    from . import commands
    cmdobj = commands.get_cmd_object(cmdname)

    doc = getdoc(cmdobj)
    if doc is None:
        raise NotImplementedError(
            "sorry, no detailed shellcomplete yet for %r" % cmdname)

    shellcomplete_on_options(cmdobj.options().values(), outfile=outfile)
    for aname in cmdobj.takes_args:
        outfile.write(aname + '\n')


def shellcomplete_on_options(options, outfile=None):
    for opt in options:
        short_name = opt.short_name()
        if short_name:
            outfile.write('"(--%s -%s)"{--%s,-%s}\n'
                          % (opt.name, short_name, opt.name, short_name))
        else:
            outfile.write('--%s\n' % opt.name)


def shellcomplete_commands(outfile=None):
    """List all commands"""
    from inspect import getdoc

    from . import commands

    commands.install_bzr_command_hooks()

    if outfile is None:
        outfile = sys.stdout

    cmds = []
    for cmdname in commands.all_command_names():
        cmd = commands.get_cmd_object(cmdname)
        cmds.append((cmdname, cmd))
        for alias in cmd.aliases:
            cmds.append((alias, cmd))
    cmds.sort()
    for cmdname, cmd in cmds:
        if cmd.hidden:
            continue
        doc = getdoc(cmd)
        if doc is None:
            outfile.write(cmdname + '\n')
        else:
            doclines = doc.splitlines()
            firstline = doclines[0].lower()
            outfile.write(cmdname + ':' + firstline[0:-1] + '\n')



#    Copyright (C) 2011 Canonical Ltd
#    Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
#    Breezy is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Breezy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Breezy; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Tests for the merge_quilt code."""

import os
import shutil

from .... import bedding, config, errors, trace
from ....merge import Merger
from ....mutabletree import MutableTree
from ....tests import TestCaseWithTransport, TestSkipped
from .. import (post_build_tree_quilt, post_merge_quilt_cleanup,
                pre_merge_quilt, start_commit_check_quilt)
from ..merge import tree_unapply_patches
from ..quilt import QuiltPatches
from . import quilt_feature

TRIVIAL_PATCH = """--- /dev/null	2012-01-02 01:09:10.986490031 +0100
+++ base/a	2012-01-02 20:03:59.710666215 +0100
@@ -0,0 +1 @@
+a
"""


def quilt_push_all(tree):
    QuiltPatches(tree, 'debian/patches').push_all()


class TestTreeUnapplyPatches(TestCaseWithTransport):

    _test_needs_features = [quilt_feature]

    def test_no_patches(self):
        tree = self.make_branch_and_tree('.')
        new_tree, target_dir = tree_unapply_patches(tree)
        self.assertIs(tree, new_tree)
        self.assertIs(None, target_dir)

    def test_unapply(self):
        orig_tree = self.make_branch_and_tree('source')
        self.build_tree(["source/debian/", "source/debian/patches/"])
        self.build_tree_contents([
            ("source/debian/patches/series", "patch1.diff\n"),
            ("source/debian/patches/patch1.diff", TRIVIAL_PATCH)])
        quilt_push_all(orig_tree)
        orig_tree.smart_add([orig_tree.basedir])
        tree, target_dir = tree_unapply_patches(orig_tree)
        self.addCleanup(shutil.rmtree, target_dir)
        self.assertPathExists("source/a")
        self.assertNotEqual(tree.basedir, orig_tree.basedir)
        self.assertPathDoesNotExist(tree.abspath("a"))
        self.assertPathExists(tree.abspath("debian/patches/series"))

    def test_unapply_nothing_applied(self):
        orig_tree = self.make_branch_and_tree('source')
        self.build_tree(["source/debian/", "source/debian/patches/"])
        self.build_tree_contents([
            ("source/debian/patches/series", "patch1.diff\n"),
            ("source/debian/patches/patch1.diff", TRIVIAL_PATCH)])
        orig_tree.smart_add([orig_tree.basedir])
        tree, target_dir = tree_unapply_patches(orig_tree)
        self.assertIs(tree, orig_tree)
        self.assertIs(None, target_dir)


class TestMergeHook(TestCaseWithTransport):

    _test_needs_features = [quilt_feature]

    def enable_hooks(self):
        Merger.hooks.install_named_hook(
            'pre_merge', pre_merge_quilt,
            'Debian quilt patch (un)applying and ancestry fixing')
        Merger.hooks.install_named_hook(
            'post_merge', post_merge_quilt_cleanup,
            'Cleaning up quilt temporary directories')
        MutableTree.hooks.install_named_hook(
            "post_build_tree", post_build_tree_quilt,
            "Apply quilt trees.")

    def test_diverged_patches(self):
        self.enable_hooks()

        tree_a = self.make_branch_and_tree('a')
        self.build_tree(
            ['a/debian/', 'a/debian/patches/', 'a/debian/source/', 'a/.pc/'])
        self.build_tree_contents([
            ('a/.pc/.quilt_patches', 'debian/patches\n'),
            ('a/.pc/.version', '2\n'),
            ('a/debian/source/format', '3.0 (quilt)'),
            ('a/debian/patches/series', 'patch1\n'),
            ('a/debian/patches/patch1', TRIVIAL_PATCH)])
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('initial')

        tree_b = tree_a.controldir.sprout('b').open_workingtree()
        self.build_tree_contents([
            ('a/debian/patches/patch1',
                "\n".join(TRIVIAL_PATCH.splitlines()[:-1] + ["+d\n"]))])
        quilt_push_all(tree_a)
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('apply patches')
        self.build_tree_contents([
            ('b/debian/patches/patch1',
                "\n".join(TRIVIAL_PATCH.splitlines()[:-1] + ["+c\n"]))])
        quilt_push_all(tree_b)
        tree_b.commit('apply patches')
        conflicts = tree_a.merge_from_branch(tree_b.branch)
        self.assertFileEqual("""\
--- /dev/null\t2012-01-02 01:09:10.986490031 +0100
+++ base/a\t2012-01-02 20:03:59.710666215 +0100
@@ -0,0 +1 @@
<<<<<<< TREE
+d
=======
+c
>>>>>>> MERGE-SOURCE
""", "a/debian/patches/patch1")
        # "a" should be unapplied again
        self.assertPathDoesNotExist("a/a")
        self.assertEqual(1, len(conflicts))

    def test_auto_apply_patches_after_checkout(self):
        self.enable_hooks()

        tree_a = self.make_branch_and_tree('a')

        self.build_tree(['a/debian/', 'a/debian/patches/'])
        self.build_tree_contents([
            ('a/debian/patches/series', 'patch1\n'),
            ('a/debian/patches/patch1', TRIVIAL_PATCH)])
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('initial')

        bedding.ensure_config_dir_exists()
        config.GlobalStack().set('quilt.tree_policy', 'applied')

        tree_a.branch.create_checkout("b")
        self.assertFileEqual("a\n", "b/a")

    def test_auto_apply_patches_after_update_format_1(self):
        self.enable_hooks()

        tree_a = self.make_branch_and_tree('a')
        tree_b = tree_a.branch.create_checkout("b")

        self.build_tree(['a/debian/', 'a/debian/patches/', 'a/.pc/'])
        self.build_tree_contents([
            ('a/.pc/.quilt_patches', 'debian/patches\n'),
            ('a/.pc/.version', '2\n'),
            ('a/debian/patches/series', 'patch1\n'),
            ('a/debian/patches/patch1', TRIVIAL_PATCH)])
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('initial')

        self.build_tree(["b/.bzr-builddeb/", "b/debian/", "b/debian/source/"])
        tree_b.get_config_stack().set('quilt.tree_policy', 'applied')
        self.build_tree_contents([
            ("b/debian/source/format", "1.0")])

        tree_b.update()
        self.assertFileEqual("a\n", "b/a")

    def test_auto_apply_patches_after_update(self):
        self.enable_hooks()

        tree_a = self.make_branch_and_tree('a')
        tree_b = tree_a.branch.create_checkout("b")

        self.build_tree(['a/debian/', 'a/debian/patches/', 'a/debian/source/', 'a/.pc/'])
        self.build_tree_contents([
            ('a/.pc/.quilt_patches', 'debian/patches\n'),
            ('a/.pc/.version', '2\n'),
            ('a/debian/source/format', '3.0 (quilt)'),
            ('a/debian/patches/series', 'patch1\n'),
            ('a/debian/patches/patch1', TRIVIAL_PATCH)])
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('initial')

        self.build_tree(["b/.bzr-builddeb/", "b/debian/", "b/debian/source/"])
        tree_b.get_config_stack().set('quilt.tree_policy', 'applied')
        self.build_tree_contents([
            ('b/debian/source/format', '3.0 (quilt)'),
            ])

        tree_b.update()
        self.assertFileEqual("a\n", "b/a")

    def test_auto_unapply_patches_after_update(self):
        self.enable_hooks()

        tree_a = self.make_branch_and_tree('a')
        tree_b = tree_a.branch.create_checkout("b")

        self.build_tree(['a/debian/', 'a/debian/patches/', 'a/debian/source/', 'a/.pc/'])
        self.build_tree_contents([
            ('a/.pc/.quilt_patches', 'debian/patches\n'),
            ('a/.pc/.version', '2\n'),
            ('a/debian/source/format', '3.0 (quilt)'),
            ('a/debian/patches/series', 'patch1\n'),
            ('a/debian/patches/patch1', TRIVIAL_PATCH)])
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('initial')

        self.build_tree(["b/.bzr-builddeb/"])
        tree_b.get_config_stack().set('quilt.tree_policy', 'unapplied')

        tree_b.update()
        self.assertPathDoesNotExist("b/a")

    def test_disabled_hook(self):
        self.enable_hooks()

        tree_a = self.make_branch_and_tree('a')
        tree_a.get_config_stack().set('quilt.smart_merge', False)
        self.build_tree(['a/debian/', 'a/debian/patches/', 'a/.pc/'])
        self.build_tree_contents([
            ('a/.pc/.quilt_patches', 'debian/patches\n'),
            ('a/.pc/.version', '2\n'),
            ('a/debian/patches/series', 'patch1\n'),
            ('a/debian/patches/patch1', TRIVIAL_PATCH),
            ("a/a", "")])
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('initial')

        tree_b = tree_a.controldir.sprout('b').open_workingtree()
        self.build_tree_contents([
            ('a/debian/patches/patch1',
                "\n".join(TRIVIAL_PATCH.splitlines()[:-1] + ["+d\n"]))])
        quilt_push_all(tree_a)
        tree_a.smart_add([tree_a.basedir])
        tree_a.commit('apply patches')
        self.assertFileEqual("d\n", "a/a")
        self.build_tree_contents([
            ('b/debian/patches/patch1',
                "\n".join(TRIVIAL_PATCH.splitlines()[:-1] + ["+c\n"]))])
        quilt_push_all(tree_b)
        tree_b.commit('apply patches')
        self.assertFileEqual("c\n", "b/a")
        conflicts = tree_a.merge_from_branch(tree_b.branch)
        self.assertFileEqual("""\
--- /dev/null\t2012-01-02 01:09:10.986490031 +0100
+++ base/a\t2012-01-02 20:03:59.710666215 +0100
@@ -0,0 +1 @@
<<<<<<< TREE
+d
=======
+c
>>>>>>> MERGE-SOURCE
""", "a/debian/patches/patch1")
        self.assertFileEqual("""\
<<<<<<< TREE
d
=======
c
>>>>>>> MERGE-SOURCE
""", "a/a")
        self.assertEqual(2, len(conflicts))



class StartCommitMergeHookTests(TestCaseWithTransport):

    _test_needs_features = [quilt_feature]

    def enable_hooks(self):
        MutableTree.hooks.install_named_hook(
            'start_commit', start_commit_check_quilt,
            'Check for (un)applied quilt patches')

    def test_applied(self):
        self.enable_hooks()
        tree = self.make_branch_and_tree('source')
        tree.get_config_stack().set('quilt.commit_policy', 'applied')
        self.build_tree(['source/debian/', 'source/debian/patches/',
                         'source/debian/source/'])
        self.build_tree_contents([
            ('source/debian/source/format', '3.0 (quilt)'),
            ('source/debian/patches/series', 'patch1\n'),
            ('source/debian/patches/patch1', TRIVIAL_PATCH)])
        self.assertPathDoesNotExist("source/.pc/applied-patches")
        self.assertPathDoesNotExist("source/a")
        tree.smart_add([tree.basedir])
        tree.commit("foo")
        self.assertPathExists("source/.pc/applied-patches")
        self.assertPathExists("source/a")

    def test_unapplied(self):
        self.enable_hooks()
        tree = self.make_branch_and_tree('source')
        tree.get_config_stack().set('quilt.commit_policy', 'unapplied')
        self.build_tree(
            ['source/debian/', 'source/debian/patches/',
             'source/debian/source/'])
        self.build_tree_contents([
            ('source/debian/patches/series', 'patch1\n'),
            ('source/debian/patches/patch1', TRIVIAL_PATCH),
            ('source/debian/source/format', '3.0 (quilt)')])
        quilt_push_all(tree)
        self.assertPathExists("source/.pc/applied-patches")
        self.assertPathExists("source/a")
        tree.smart_add([tree.basedir])
        tree.commit("foo")
        self.assertPathDoesNotExist("source/.pc/applied-patches")
        self.assertPathDoesNotExist("source/a")

    def test_warning(self):
        self.enable_hooks()
        warnings = []

        def warning(*args):
            if len(args) > 1:
                warnings.append(args[0] % args[1:])
            else:
                warnings.append(args[0])
        _warning = trace.warning
        trace.warning = warning
        self.addCleanup(setattr, trace, "warning", _warning)
        tree = self.make_branch_and_tree('source')
        self.build_tree(['source/debian/', 'source/debian/patches/',
                         'source/debian/source/'])
        self.build_tree_contents([
            ('source/debian/patches/series', 'patch1\n'),
            ('source/debian/patches/patch1', TRIVIAL_PATCH)])
        quilt_push_all(tree)
        tree.smart_add([tree.basedir])
        tree.commit("initial")
        self.assertEqual([], warnings)
        self.assertPathExists("source/.pc/applied-patches")
        self.assertPathExists("source/a")
        self.build_tree_contents([
            ('source/debian/source/format', '3.0 (quilt)'),
            ('source/debian/patches/series', 'patch1\npatch2\n'),
            ('source/debian/patches/patch2',
                """--- /dev/null	2012-01-02 01:09:10.986490031 +0100
+++ base/b	2012-01-02 20:03:59.710666215 +0100
@@ -0,0 +1 @@
+a
""")])
        tree.smart_add([tree.basedir])
        tree.commit("foo")
        self.assertEqual(
            ['Committing with 1 patches applied and 1 patches unapplied.'],
            warnings)
        self.assertPathExists("source/.pc/applied-patches")
        self.assertPathExists("source/a")
        self.assertPathDoesNotExist("source/b")



# Copyright (C) 2008-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for group compression."""

import zlib

from ... import config, errors, osutils, tests, trace
from ...osutils import sha_string
from ...tests.scenarios import load_tests_apply_scenarios
from .. import btree_index, groupcompress
from .. import index as _mod_index
from .. import knit, versionedfile
from .test__groupcompress import compiled_groupcompress_feature


def group_compress_implementation_scenarios():
    scenarios = [
        ('python', {'compressor': groupcompress.PythonGroupCompressor}),
        ]
    if compiled_groupcompress_feature.available():
        scenarios.append(('C',
                          {'compressor': groupcompress.PyrexGroupCompressor}))
    return scenarios


load_tests = load_tests_apply_scenarios


class TestGroupCompressor(tests.TestCase):

    def _chunks_to_repr_lines(self, chunks):
        return '\n'.join(map(repr, b''.join(chunks).split(b'\n')))

    def assertEqualDiffEncoded(self, expected, actual):
        """Compare the actual content to the expected content.

        :param expected: A group of chunks that we expect to see
        :param actual: The measured 'chunks'

        We will transform the chunks back into lines, and then run 'repr()'
        over them to handle non-ascii characters.
        """
        self.assertEqualDiff(self._chunks_to_repr_lines(expected),
                             self._chunks_to_repr_lines(actual))


class TestAllGroupCompressors(TestGroupCompressor):
    """Tests for GroupCompressor"""

    scenarios = group_compress_implementation_scenarios()
    compressor = None  # Set by scenario

    def test_empty_delta(self):
        compressor = self.compressor()
        self.assertEqual([], compressor.chunks)

    def test_one_nosha_delta(self):
        # diff against NUKK
        compressor = self.compressor()
        text = b'strange\ncommon\n'
        sha1, start_point, end_point, _ = compressor.compress(
            ('label',), [text], len(text), None)
        self.assertEqual(sha_string(b'strange\ncommon\n'), sha1)
        expected_lines = b'f\x0fstrange\ncommon\n'
        self.assertEqual(expected_lines, b''.join(compressor.chunks))
        self.assertEqual(0, start_point)
        self.assertEqual(len(expected_lines), end_point)

    def test_empty_content(self):
        compressor = self.compressor()
        # Adding empty bytes should return the 'null' record
        sha1, start_point, end_point, kind = compressor.compress(
            ('empty',), [], 0, None)
        self.assertEqual(0, start_point)
        self.assertEqual(0, end_point)
        self.assertEqual('fulltext', kind)
        self.assertEqual(groupcompress._null_sha1, sha1)
        self.assertEqual(0, compressor.endpoint)
        self.assertEqual([], compressor.chunks)
        # Even after adding some content
        text = b'some\nbytes\n'
        compressor.compress(('content',), [text], len(text), None)
        self.assertTrue(compressor.endpoint > 0)
        sha1, start_point, end_point, kind = compressor.compress(
            ('empty2',), [], 0, None)
        self.assertEqual(0, start_point)
        self.assertEqual(0, end_point)
        self.assertEqual('fulltext', kind)
        self.assertEqual(groupcompress._null_sha1, sha1)

    def test_extract_from_compressor(self):
        # Knit fetching will try to reconstruct texts locally which results in
        # reading something that is in the compressor stream already.
        compressor = self.compressor()
        text = b'strange\ncommon long line\nthat needs a 16 byte match\n'
        sha1_1, _, _, _ = compressor.compress(
            ('label',), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = b'common long line\nthat needs a 16 byte match\ndifferent\n'
        sha1_2, _, end_point, _ = compressor.compress(
            ('newlabel',), [text], len(text), None)
        # get the first out
        self.assertEqual(([b'strange\ncommon long line\n'
                           b'that needs a 16 byte match\n'], sha1_1),
                         compressor.extract(('label',)))
        # and the second
        self.assertEqual(([b'common long line\nthat needs a 16 byte match\n'
                           b'different\n'], sha1_2),
                         compressor.extract(('newlabel',)))

    def test_pop_last(self):
        compressor = self.compressor()
        text = b'some text\nfor the first entry\n'
        _, _, _, _ = compressor.compress(
            ('key1',), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = b'some text\nfor the second entry\n'
        _, _, _, _ = compressor.compress(
            ('key2',), [text], len(text), None)
        compressor.pop_last()
        self.assertEqual(expected_lines, compressor.chunks)


class TestPyrexGroupCompressor(TestGroupCompressor):

    _test_needs_features = [compiled_groupcompress_feature]
    compressor = groupcompress.PyrexGroupCompressor

    def test_stats(self):
        compressor = self.compressor()
        chunks = [b'strange\n',
                  b'common very very long line\n',
                  b'plus more text\n']
        compressor.compress(
            ('label',), chunks, sum(map(len, chunks)), None)
        chunks = [
            b'common very very long line\n',
            b'plus more text\n',
            b'different\n',
            b'moredifferent\n']
        compressor.compress(
            ('newlabel',),
            chunks, sum(map(len, chunks)), None)
        chunks = [
            b'new\n',
            b'common very very long line\n',
            b'plus more text\n',
            b'different\n',
            b'moredifferent\n']
        compressor.compress(
            ('label3',), chunks, sum(map(len, chunks)), None)
        self.assertAlmostEqual(1.9, compressor.ratio(), 1)

    def test_two_nosha_delta(self):
        compressor = self.compressor()
        text = b'strange\ncommon long line\nthat needs a 16 byte match\n'
        sha1_1, _, _, _ = compressor.compress(('label',), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = b'common long line\nthat needs a 16 byte match\ndifferent\n'
        sha1_2, start_point, end_point, _ = compressor.compress(
            ('newlabel',), [text], len(text), None)
        self.assertEqual(sha_string(text), sha1_2)
        expected_lines.extend([
            # 'delta', delta length
            b'd\x0f',
            # source and target length
            b'\x36',
            # copy the line common
            b'\x91\x0a\x2c',  # copy, offset 0x0a, len 0x2c
            # add the line different, and the trailing newline
            b'\x0adifferent\n',  # insert 10 bytes
            ])
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)

    def test_three_nosha_delta(self):
        # The first interesting test: make a change that should use lines from
        # both parents.
        compressor = self.compressor()
        text = b'strange\ncommon very very long line\nwith some extra text\n'
        sha1_1, _, _, _ = compressor.compress(
            ('label',), [text], len(text), None)
        text = b'different\nmoredifferent\nand then some more\n'
        sha1_2, _, _, _ = compressor.compress(
            ('newlabel',), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = (b'new\ncommon very very long line\nwith some extra text\n'
                b'different\nmoredifferent\nand then some more\n')
        sha1_3, start_point, end_point, _ = compressor.compress(
            ('label3',), [text], len(text), None)
        self.assertEqual(sha_string(text), sha1_3)
        expected_lines.extend([
            # 'delta', delta length
            b'd\x0b',
            # source and target length
            b'\x5f'
            # insert new
            b'\x03new',
            # Copy of first parent 'common' range
            b'\x91\x09\x31'  # copy, offset 0x09, 0x31 bytes
            # Copy of second parent 'different' range
            b'\x91\x3c\x2b'  # copy, offset 0x3c, 0x2b bytes
            ])
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)


class TestPythonGroupCompressor(TestGroupCompressor):

    compressor = groupcompress.PythonGroupCompressor

    def test_stats(self):
        compressor = self.compressor()
        chunks = [b'strange\n',
                  b'common very very long line\n',
                  b'plus more text\n']
        compressor.compress(
            ('label',), chunks, sum(map(len, chunks)), None)
        chunks = [
            b'common very very long line\n',
            b'plus more text\n',
            b'different\n',
            b'moredifferent\n']
        compressor.compress(
            ('newlabel',), chunks, sum(map(len, chunks)), None)
        chunks = [
            b'new\n',
            b'common very very long line\n',
            b'plus more text\n',
            b'different\n',
            b'moredifferent\n']
        compressor.compress(
            ('label3',),
            chunks, sum(map(len, chunks)), None)
        self.assertAlmostEqual(1.9, compressor.ratio(), 1)

    def test_two_nosha_delta(self):
        compressor = self.compressor()
        text = b'strange\ncommon long line\nthat needs a 16 byte match\n'
        sha1_1, _, _, _ = compressor.compress(
            ('label',), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = b'common long line\nthat needs a 16 byte match\ndifferent\n'
        sha1_2, start_point, end_point, _ = compressor.compress(
            ('newlabel',), [text], len(text), None)
        self.assertEqual(sha_string(text), sha1_2)
        expected_lines.extend([
            # 'delta', delta length
            b'd\x0f',
            # target length
            b'\x36',
            # copy the line common
            b'\x91\x0a\x2c',  # copy, offset 0x0a, len 0x2c
            # add the line different, and the trailing newline
            b'\x0adifferent\n',  # insert 10 bytes
            ])
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)

    def test_three_nosha_delta(self):
        # The first interesting test: make a change that should use lines from
        # both parents.
        compressor = self.compressor()
        text = b'strange\ncommon very very long line\nwith some extra text\n'
        sha1_1, _, _, _ = compressor.compress(
            ('label',), [text], len(text), None)
        text = b'different\nmoredifferent\nand then some more\n'
        sha1_2, _, _, _ = compressor.compress(
            ('newlabel',), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = (b'new\ncommon very very long line\nwith some extra text\n'
                b'different\nmoredifferent\nand then some more\n')
        sha1_3, start_point, end_point, _ = compressor.compress(
            ('label3',), [text], len(text), None)
        self.assertEqual(sha_string(text), sha1_3)
        expected_lines.extend([
            # 'delta', delta length
            b'd\x0c',
            # target length
            b'\x5f'
            # insert new
            b'\x04new\n',
            # Copy of first parent 'common' range
            b'\x91\x0a\x30'  # copy, offset 0x0a, 0x30 bytes
            # Copy of second parent 'different' range
            b'\x91\x3c\x2b'  # copy, offset 0x3c, 0x2b bytes
            ])
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)


class TestGroupCompressBlock(tests.TestCase):

    def make_block(self, key_to_text):
        """Create a GroupCompressBlock, filling it with the given texts."""
        compressor = groupcompress.GroupCompressor()
        start = 0
        for key in sorted(key_to_text):
            compressor.compress(
                key, [key_to_text[key]], len(key_to_text[key]), None)
        locs = {key: (start, end) for key, (start, _, end, _)
                    in compressor.labels_deltas.items()}
        block = compressor.flush()
        raw_bytes = block.to_bytes()
        # Go through from_bytes(to_bytes()) so that we start with a compressed
        # content object
        return locs, groupcompress.GroupCompressBlock.from_bytes(raw_bytes)

    def test_from_empty_bytes(self):
        self.assertRaises(ValueError,
                          groupcompress.GroupCompressBlock.from_bytes, b'')

    def test_from_minimal_bytes(self):
        block = groupcompress.GroupCompressBlock.from_bytes(
            b'gcb1z\n0\n0\n')
        self.assertIsInstance(block, groupcompress.GroupCompressBlock)
        self.assertIs(None, block._content)
        self.assertEqual(b'', block._z_content)
        block._ensure_content()
        self.assertEqual(b'', block._content)
        self.assertEqual(b'', block._z_content)
        block._ensure_content()  # Ensure content is safe to call 2x

    def test_from_invalid(self):
        self.assertRaises(ValueError,
                          groupcompress.GroupCompressBlock.from_bytes,
                          b'this is not a valid header')

    def test_from_bytes(self):
        content = (b'a tiny bit of content\n')
        z_content = zlib.compress(content)
        z_bytes = (
            b'gcb1z\n'  # group compress block v1 plain
            b'%d\n'  # Length of compressed content
            b'%d\n'  # Length of uncompressed content
            b'%s'   # Compressed content
            ) % (len(z_content), len(content), z_content)
        block = groupcompress.GroupCompressBlock.from_bytes(
            z_bytes)
        self.assertEqual(z_content, block._z_content)
        self.assertIs(None, block._content)
        self.assertEqual(len(z_content), block._z_content_length)
        self.assertEqual(len(content), block._content_length)
        block._ensure_content()
        self.assertEqual(z_content, block._z_content)
        self.assertEqual(content, block._content)

    def test_to_chunks(self):
        content_chunks = [b'this is some content\n',
                          b'this content will be compressed\n']
        content_len = sum(map(len, content_chunks))
        content = b''.join(content_chunks)
        gcb = groupcompress.GroupCompressBlock()
        gcb.set_chunked_content(content_chunks, content_len)
        total_len, block_chunks = gcb.to_chunks()
        block_bytes = b''.join(block_chunks)
        self.assertEqual(gcb._z_content_length, len(gcb._z_content))
        self.assertEqual(total_len, len(block_bytes))
        self.assertEqual(gcb._content_length, content_len)
        expected_header = (b'gcb1z\n'  # group compress block v1 zlib
                           b'%d\n'  # Length of compressed content
                           b'%d\n'  # Length of uncompressed content
                           ) % (gcb._z_content_length, gcb._content_length)
        # The first chunk should be the header chunk. It is small, fixed size,
        # and there is no compelling reason to split it up
        self.assertEqual(expected_header, block_chunks[0])
        self.assertStartsWith(block_bytes, expected_header)
        remaining_bytes = block_bytes[len(expected_header):]
        raw_bytes = zlib.decompress(remaining_bytes)
        self.assertEqual(content, raw_bytes)

    def test_to_bytes(self):
        content = (b'this is some content\n'
                   b'this content will be compressed\n')
        gcb = groupcompress.GroupCompressBlock()
        gcb.set_content(content)
        data = gcb.to_bytes()
        self.assertEqual(gcb._z_content_length, len(gcb._z_content))
        self.assertEqual(gcb._content_length, len(content))
        expected_header = (b'gcb1z\n'  # group compress block v1 zlib
                           b'%d\n'  # Length of compressed content
                           b'%d\n'  # Length of uncompressed content
                           ) % (gcb._z_content_length, gcb._content_length)
        self.assertStartsWith(data, expected_header)
        remaining_bytes = data[len(expected_header):]
        raw_bytes = zlib.decompress(remaining_bytes)
        self.assertEqual(content, raw_bytes)

        # we should get the same results if using the chunked version
        gcb = groupcompress.GroupCompressBlock()
        gcb.set_chunked_content([b'this is some content\n'
                                 b'this content will be compressed\n'],
                                len(content))
        old_data = data
        data = gcb.to_bytes()
        self.assertEqual(old_data, data)

    def test_partial_decomp(self):
        content_chunks = []
        # We need a sufficient amount of data so that zlib.decompress has
        # partial decompression to work with. Most auto-generated data
        # compresses a bit too well, we want a combination, so we combine a sha
        # hash with compressible data.
        for i in range(2048):
            next_content = b'%d\nThis is a bit of duplicate text\n' % (i,)
            content_chunks.append(next_content)
            next_sha1 = osutils.sha_string(next_content)
            content_chunks.append(next_sha1 + b'\n')
        content = b''.join(content_chunks)
        self.assertEqual(158634, len(content))
        z_content = zlib.compress(content)
        self.assertEqual(57182, len(z_content))
        block = groupcompress.GroupCompressBlock()
        block._z_content_chunks = (z_content,)
        block._z_content_length = len(z_content)
        block._compressor_name = 'zlib'
        block._content_length = 158634
        self.assertIs(None, block._content)
        block._ensure_content(100)
        self.assertIsNot(None, block._content)
        # We have decompressed at least 100 bytes
        self.assertTrue(len(block._content) >= 100)
        # We have not decompressed the whole content
        self.assertTrue(len(block._content) < 158634)
        self.assertEqualDiff(content[:len(block._content)], block._content)
        # ensuring content that we already have shouldn't cause any more data
        # to be extracted
        cur_len = len(block._content)
        block._ensure_content(cur_len - 10)
        self.assertEqual(cur_len, len(block._content))
        # Now we want a bit more content
        cur_len += 10
        block._ensure_content(cur_len)
        self.assertTrue(len(block._content) >= cur_len)
        self.assertTrue(len(block._content) < 158634)
        self.assertEqualDiff(content[:len(block._content)], block._content)
        # And now lets finish
        block._ensure_content(158634)
        self.assertEqualDiff(content, block._content)
        # And the decompressor is finalized
        self.assertIs(None, block._z_content_decompressor)

    def test__ensure_all_content(self):
        content_chunks = []
        # We need a sufficient amount of data so that zlib.decompress has
        # partial decompression to work with. Most auto-generated data
        # compresses a bit too well, we want a combination, so we combine a sha
        # hash with compressible data.
        for i in range(2048):
            next_content = b'%d\nThis is a bit of duplicate text\n' % (i,)
            content_chunks.append(next_content)
            next_sha1 = osutils.sha_string(next_content)
            content_chunks.append(next_sha1 + b'\n')
        content = b''.join(content_chunks)
        self.assertEqual(158634, len(content))
        z_content = zlib.compress(content)
        self.assertEqual(57182, len(z_content))
        block = groupcompress.GroupCompressBlock()
        block._z_content_chunks = (z_content,)
        block._z_content_length = len(z_content)
        block._compressor_name = 'zlib'
        block._content_length = 158634
        self.assertIs(None, block._content)
        # The first _ensure_content got all of the required data
        block._ensure_content(158634)
        self.assertEqualDiff(content, block._content)
        # And we should have released the _z_content_decompressor since it was
        # fully consumed
        self.assertIs(None, block._z_content_decompressor)

    def test__dump(self):
        dup_content = b'some duplicate content\nwhich is sufficiently long\n'
        key_to_text = {(b'1',): dup_content + b'1 unique\n',
                       (b'2',): dup_content + b'2 extra special\n'}
        locs, block = self.make_block(key_to_text)
        self.assertEqual([(b'f', len(key_to_text[(b'1',)])),
                          (b'd', 21, len(key_to_text[(b'2',)]),
                           [(b'c', 2, len(dup_content)),
                            (b'i', len(b'2 extra special\n'), b'')
                            ]),
                          ], block._dump())


class TestCaseWithGroupCompressVersionedFiles(
        tests.TestCaseWithMemoryTransport):

    def make_test_vf(self, create_graph, keylength=1, do_cleanup=True,
                     dir='.', inconsistency_fatal=True):
        t = self.get_transport(dir)
        t.ensure_base()
        vf = groupcompress.make_pack_factory(graph=create_graph,
                                             delta=False, keylength=keylength,
                                             inconsistency_fatal=inconsistency_fatal)(t)
        if do_cleanup:
            self.addCleanup(groupcompress.cleanup_pack_group, vf)
        return vf


class TestGroupCompressVersionedFiles(TestCaseWithGroupCompressVersionedFiles):

    def make_g_index(self, name, ref_lists=0, nodes=[]):
        builder = btree_index.BTreeBuilder(ref_lists)
        for node, references, value in nodes:
            builder.add_node(node, references, value)
        stream = builder.finish()
        trans = self.get_transport()
        size = trans.put_file(name, stream)
        return btree_index.BTreeGraphIndex(trans, name, size)

    def make_g_index_missing_parent(self):
        graph_index = self.make_g_index('missing_parent', 1,
                                        [((b'parent', ), b'2 78 2 10', ([],)),
                                         ((b'tip', ), b'2 78 2 10',
                                            ([(b'parent', ), (b'missing-parent', )],)),
                                         ])
        return graph_index

    def test_get_record_stream_as_requested(self):
        # Consider promoting 'as-requested' to general availability, and
        # make this a VF interface test
        vf = self.make_test_vf(False, dir='source')
        vf.add_lines((b'a',), (), [b'lines\n'])
        vf.add_lines((b'b',), (), [b'lines\n'])
        vf.add_lines((b'c',), (), [b'lines\n'])
        vf.add_lines((b'd',), (), [b'lines\n'])
        vf.writer.end()
        keys = [record.key for record in vf.get_record_stream(
            [(b'a',), (b'b',), (b'c',), (b'd',)],
            'as-requested', False)]
        self.assertEqual([(b'a',), (b'b',), (b'c',), (b'd',)], keys)
        keys = [record.key for record in vf.get_record_stream(
            [(b'b',), (b'a',), (b'd',), (b'c',)],
            'as-requested', False)]
        self.assertEqual([(b'b',), (b'a',), (b'd',), (b'c',)], keys)

        # It should work even after being repacked into another VF
        vf2 = self.make_test_vf(False, dir='target')
        vf2.insert_record_stream(vf.get_record_stream(
            [(b'b',), (b'a',), (b'd',), (b'c',)], 'as-requested', False))
        vf2.writer.end()

        keys = [record.key for record in vf2.get_record_stream(
            [(b'a',), (b'b',), (b'c',), (b'd',)],
            'as-requested', False)]
        self.assertEqual([(b'a',), (b'b',), (b'c',), (b'd',)], keys)
        keys = [record.key for record in vf2.get_record_stream(
            [(b'b',), (b'a',), (b'd',), (b'c',)],
            'as-requested', False)]
        self.assertEqual([(b'b',), (b'a',), (b'd',), (b'c',)], keys)

    def test_get_record_stream_max_bytes_to_index_default(self):
        vf = self.make_test_vf(True, dir='source')
        vf.add_lines((b'a',), (), [b'lines\n'])
        vf.writer.end()
        record = next(vf.get_record_stream([(b'a',)], 'unordered', True))
        self.assertEqual(vf._DEFAULT_COMPRESSOR_SETTINGS,
                         record._manager._get_compressor_settings())

    def test_get_record_stream_accesses_compressor_settings(self):
        vf = self.make_test_vf(True, dir='source')
        vf.add_lines((b'a',), (), [b'lines\n'])
        vf.writer.end()
        vf._max_bytes_to_index = 1234
        record = next(vf.get_record_stream([(b'a',)], 'unordered', True))
        self.assertEqual(dict(max_bytes_to_index=1234),
                         record._manager._get_compressor_settings())

    @staticmethod
    def grouped_stream(revision_ids, first_parents=()):
        parents = first_parents
        for revision_id in revision_ids:
            key = (revision_id,)
            record = versionedfile.FulltextContentFactory(
                key, parents, None,
                b'some content that is\n'
                b'identical except for\n'
                b'revision_id:%s\n' % (revision_id,))
            yield record
            parents = (key,)

    def test_insert_record_stream_reuses_blocks(self):
        vf = self.make_test_vf(True, dir='source')
        # One group, a-d
        vf.insert_record_stream(self.grouped_stream([b'a', b'b', b'c', b'd']))
        # Second group, e-h
        vf.insert_record_stream(self.grouped_stream(
            [b'e', b'f', b'g', b'h'], first_parents=((b'd',),)))
        block_bytes = {}
        stream = vf.get_record_stream(
            [(r.encode(),) for r in 'abcdefgh'], 'unordered', False)
        num_records = 0
        for record in stream:
            if record.key in [(b'a',), (b'e',)]:
                self.assertEqual('groupcompress-block', record.storage_kind)
            else:
                self.assertEqual('groupcompress-block-ref',
                                 record.storage_kind)
            block_bytes[record.key] = record._manager._block._z_content
            num_records += 1
        self.assertEqual(8, num_records)
        for r in 'abcd':
            key = (r.encode(),)
            self.assertIs(block_bytes[key], block_bytes[(b'a',)])
            self.assertNotEqual(block_bytes[key], block_bytes[(b'e',)])
        for r in 'efgh':
            key = (r.encode(),)
            self.assertIs(block_bytes[key], block_bytes[(b'e',)])
            self.assertNotEqual(block_bytes[key], block_bytes[(b'a',)])
        # Now copy the blocks into another vf, and ensure that the blocks are
        # preserved without creating new entries
        vf2 = self.make_test_vf(True, dir='target')
        keys = [(r.encode(),) for r in 'abcdefgh']
        # ordering in 'groupcompress' order, should actually swap the groups in
        # the target vf, but the groups themselves should not be disturbed.

        def small_size_stream():
            for record in vf.get_record_stream(keys, 'groupcompress', False):
                record._manager._full_enough_block_size = \
                    record._manager._block._content_length
                yield record

        vf2.insert_record_stream(small_size_stream())
        stream = vf2.get_record_stream(keys, 'groupcompress', False)
        vf2.writer.end()
        num_records = 0
        for record in stream:
            num_records += 1
            self.assertEqual(block_bytes[record.key],
                             record._manager._block._z_content)
        self.assertEqual(8, num_records)

    def test_insert_record_stream_packs_on_the_fly(self):
        vf = self.make_test_vf(True, dir='source')
        # One group, a-d
        vf.insert_record_stream(self.grouped_stream([b'a', b'b', b'c', b'd']))
        # Second group, e-h
        vf.insert_record_stream(self.grouped_stream(
            [b'e', b'f', b'g', b'h'], first_parents=((b'd',),)))
        # Now copy the blocks into another vf, and see that the
        # insert_record_stream rebuilt a new block on-the-fly because of
        # under-utilization
        vf2 = self.make_test_vf(True, dir='target')
        keys = [(r.encode(),) for r in 'abcdefgh']
        vf2.insert_record_stream(vf.get_record_stream(
            keys, 'groupcompress', False))
        stream = vf2.get_record_stream(keys, 'groupcompress', False)
        vf2.writer.end()
        num_records = 0
        # All of the records should be recombined into a single block
        block = None
        for record in stream:
            num_records += 1
            if block is None:
                block = record._manager._block
            else:
                self.assertIs(block, record._manager._block)
        self.assertEqual(8, num_records)

    def test__insert_record_stream_no_reuse_block(self):
        vf = self.make_test_vf(True, dir='source')
        # One group, a-d
        vf.insert_record_stream(self.grouped_stream([b'a', b'b', b'c', b'd']))
        # Second group, e-h
        vf.insert_record_stream(self.grouped_stream(
            [b'e', b'f', b'g', b'h'], first_parents=((b'd',),)))
        vf.writer.end()
        keys = [(r.encode(),) for r in 'abcdefgh']
        self.assertEqual(8, len(list(
            vf.get_record_stream(keys, 'unordered', False))))
        # Now copy the blocks into another vf, and ensure that the blocks are
        # preserved without creating new entries
        vf2 = self.make_test_vf(True, dir='target')
        # ordering in 'groupcompress' order, should actually swap the groups in
        # the target vf, but the groups themselves should not be disturbed.
        list(vf2._insert_record_stream(vf.get_record_stream(
            keys, 'groupcompress', False),
            reuse_blocks=False))
        vf2.writer.end()
        # After inserting with reuse_blocks=False, we should have everything in
        # a single new block.
        stream = vf2.get_record_stream(keys, 'groupcompress', False)
        block = None
        for record in stream:
            if block is None:
                block = record._manager._block
            else:
                self.assertIs(block, record._manager._block)

    def test_add_missing_noncompression_parent_unvalidated_index(self):
        unvalidated = self.make_g_index_missing_parent()
        combined = _mod_index.CombinedGraphIndex([unvalidated])
        index = groupcompress._GCGraphIndex(combined,
                                            is_locked=lambda: True, parents=True,
                                            track_external_parent_refs=True)
        index.scan_unvalidated_index(unvalidated)
        self.assertEqual(
            frozenset([(b'missing-parent',)]), index.get_missing_parents())

    def test_track_external_parent_refs(self):
        g_index = self.make_g_index('empty', 1, [])
        mod_index = btree_index.BTreeBuilder(1, 1)
        combined = _mod_index.CombinedGraphIndex([g_index, mod_index])
        index = groupcompress._GCGraphIndex(combined,
                                            is_locked=lambda: True, parents=True,
                                            add_callback=mod_index.add_nodes,
                                            track_external_parent_refs=True)
        index.add_records([
            ((b'new-key',), b'2 10 2 10', [((b'parent-1',), (b'parent-2',))])])
        self.assertEqual(
            frozenset([(b'parent-1',), (b'parent-2',)]),
            index.get_missing_parents())

    def make_source_with_b(self, a_parent, path):
        source = self.make_test_vf(True, dir=path)
        source.add_lines((b'a',), (), [b'lines\n'])
        if a_parent:
            b_parents = ((b'a',),)
        else:
            b_parents = ()
        source.add_lines((b'b',), b_parents, [b'lines\n'])
        return source

    def do_inconsistent_inserts(self, inconsistency_fatal):
        target = self.make_test_vf(True, dir='target',
                                   inconsistency_fatal=inconsistency_fatal)
        for x in range(2):
            source = self.make_source_with_b(x == 1, 'source%s' % x)
            target.insert_record_stream(source.get_record_stream(
                [(b'b',)], 'unordered', False))

    def test_inconsistent_redundant_inserts_warn(self):
        """Should not insert a record that is already present."""
        warnings = []

        def warning(template, args):
            warnings.append(template % args)
        _trace_warning = trace.warning
        trace.warning = warning
        try:
            self.do_inconsistent_inserts(inconsistency_fatal=False)
        finally:
            trace.warning = _trace_warning
        self.assertContainsRe(
            "\n".join(warnings),
            r"^inconsistent details in skipped record: \(b?'b',\)"
            r" \(b?'42 32 0 8', \(\(\),\)\)"
            r" \(b?'74 32 0 8', \(\(\(b?'a',\),\),\)\)$")

    def test_inconsistent_redundant_inserts_raises(self):
        e = self.assertRaises(knit.KnitCorrupt, self.do_inconsistent_inserts,
                              inconsistency_fatal=True)
        self.assertContainsRe(str(e), r"Knit.* corrupt: inconsistent details"
                              r" in add_records:"
                              r" \(b?'b',\) \(b?'42 32 0 8', \(\(\),\)\)"
                              r" \(b?'74 32 0 8', \(\(\(b?'a',\),\),\)\)")

    def test_clear_cache(self):
        vf = self.make_source_with_b(True, 'source')
        vf.writer.end()
        for record in vf.get_record_stream([(b'a',), (b'b',)], 'unordered',
                                           True):
            pass
        self.assertTrue(len(vf._group_cache) > 0)
        vf.clear_cache()
        self.assertEqual(0, len(vf._group_cache))


class TestGroupCompressConfig(tests.TestCaseWithTransport):

    def make_test_vf(self):
        t = self.get_transport('.')
        t.ensure_base()
        factory = groupcompress.make_pack_factory(graph=True,
                                                  delta=False, keylength=1, inconsistency_fatal=True)
        vf = factory(t)
        self.addCleanup(groupcompress.cleanup_pack_group, vf)
        return vf

    def test_max_bytes_to_index_default(self):
        vf = self.make_test_vf()
        gc = vf._make_group_compressor()
        self.assertEqual(vf._DEFAULT_MAX_BYTES_TO_INDEX,
                         vf._max_bytes_to_index)
        if isinstance(gc, groupcompress.PyrexGroupCompressor):
            self.assertEqual(vf._DEFAULT_MAX_BYTES_TO_INDEX,
                             gc._delta_index._max_bytes_to_index)

    def test_max_bytes_to_index_in_config(self):
        c = config.GlobalConfig()
        c.set_user_option('bzr.groupcompress.max_bytes_to_index', '10000')
        vf = self.make_test_vf()
        gc = vf._make_group_compressor()
        self.assertEqual(10000, vf._max_bytes_to_index)
        if isinstance(gc, groupcompress.PyrexGroupCompressor):
            self.assertEqual(10000, gc._delta_index._max_bytes_to_index)

    def test_max_bytes_to_index_bad_config(self):
        c = config.GlobalConfig()
        c.set_user_option('bzr.groupcompress.max_bytes_to_index', 'boogah')
        vf = self.make_test_vf()
        # TODO: This is triggering a warning, we might want to trap and make
        #       sure it is readable.
        gc = vf._make_group_compressor()
        self.assertEqual(vf._DEFAULT_MAX_BYTES_TO_INDEX,
                         vf._max_bytes_to_index)
        if isinstance(gc, groupcompress.PyrexGroupCompressor):
            self.assertEqual(vf._DEFAULT_MAX_BYTES_TO_INDEX,
                             gc._delta_index._max_bytes_to_index)


class StubGCVF:
    def __init__(self, canned_get_blocks=None):
        self._group_cache = {}
        self._canned_get_blocks = canned_get_blocks or []

    def _get_blocks(self, read_memos):
        return iter(self._canned_get_blocks)


class Test_BatchingBlockFetcher(TestCaseWithGroupCompressVersionedFiles):
    """Simple whitebox unit tests for _BatchingBlockFetcher."""

    def test_add_key_new_read_memo(self):
        """Adding a key with an uncached read_memo new to this batch adds that
        read_memo to the list of memos to fetch.
        """
        # locations are: index_memo, ignored, parents, ignored
        # where index_memo is: (idx, offset, len, factory_start, factory_end)
        # and (idx, offset, size) is known as the 'read_memo', identifying the
        # raw bytes needed.
        read_memo = ('fake index', 100, 50)
        locations = {
            ('key',): (read_memo + (None, None), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(StubGCVF(), locations)
        total_size = batcher.add_key(('key',))
        self.assertEqual(50, total_size)
        self.assertEqual([('key',)], batcher.keys)
        self.assertEqual([read_memo], batcher.memos_to_get)

    def test_add_key_duplicate_read_memo(self):
        """read_memos that occur multiple times in a batch will only be fetched
        once.
        """
        read_memo = ('fake index', 100, 50)
        # Two keys, both sharing the same read memo (but different overall
        # index_memos).
        locations = {
            ('key1',): (read_memo + (0, 1), None, None, None),
            ('key2',): (read_memo + (1, 2), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(StubGCVF(), locations)
        total_size = batcher.add_key(('key1',))
        total_size = batcher.add_key(('key2',))
        self.assertEqual(50, total_size)
        self.assertEqual([('key1',), ('key2',)], batcher.keys)
        self.assertEqual([read_memo], batcher.memos_to_get)

    def test_add_key_cached_read_memo(self):
        """Adding a key with a cached read_memo will not cause that read_memo
        to be added to the list to fetch.
        """
        read_memo = ('fake index', 100, 50)
        gcvf = StubGCVF()
        gcvf._group_cache[read_memo] = 'fake block'
        locations = {
            ('key',): (read_memo + (None, None), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(gcvf, locations)
        total_size = batcher.add_key(('key',))
        self.assertEqual(0, total_size)
        self.assertEqual([('key',)], batcher.keys)
        self.assertEqual([], batcher.memos_to_get)

    def test_yield_factories_empty(self):
        """An empty batch yields no factories."""
        batcher = groupcompress._BatchingBlockFetcher(StubGCVF(), {})
        self.assertEqual([], list(batcher.yield_factories()))

    def test_yield_factories_calls_get_blocks(self):
        """Uncached memos are retrieved via get_blocks."""
        read_memo1 = ('fake index', 100, 50)
        read_memo2 = ('fake index', 150, 40)
        gcvf = StubGCVF(
            canned_get_blocks=[
                (read_memo1, groupcompress.GroupCompressBlock()),
                (read_memo2, groupcompress.GroupCompressBlock())])
        locations = {
            ('key1',): (read_memo1 + (0, 0), None, None, None),
            ('key2',): (read_memo2 + (0, 0), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(gcvf, locations)
        batcher.add_key(('key1',))
        batcher.add_key(('key2',))
        factories = list(batcher.yield_factories(full_flush=True))
        self.assertLength(2, factories)
        keys = [f.key for f in factories]
        kinds = [f.storage_kind for f in factories]
        self.assertEqual([('key1',), ('key2',)], keys)
        self.assertEqual(['groupcompress-block', 'groupcompress-block'], kinds)

    def test_yield_factories_flushing(self):
        """yield_factories holds back on yielding results from the final block
        unless passed full_flush=True.
        """
        fake_block = groupcompress.GroupCompressBlock()
        read_memo = ('fake index', 100, 50)
        gcvf = StubGCVF()
        gcvf._group_cache[read_memo] = fake_block
        locations = {
            ('key',): (read_memo + (0, 0), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(gcvf, locations)
        batcher.add_key(('key',))
        self.assertEqual([], list(batcher.yield_factories()))
        factories = list(batcher.yield_factories(full_flush=True))
        self.assertLength(1, factories)
        self.assertEqual(('key',), factories[0].key)
        self.assertEqual('groupcompress-block', factories[0].storage_kind)


class TestLazyGroupCompress(tests.TestCaseWithTransport):

    _texts = {
        (b'key1',): b"this is a text\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
        (b'key2',): b"another text\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
        (b'key3',): b"yet another text which won't be extracted\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
        (b'key4',): b"this will be extracted\n"
        b"but references most of its bytes from\n"
        b"yet another text which won't be extracted\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
    }

    def make_block(self, key_to_text):
        """Create a GroupCompressBlock, filling it with the given texts."""
        compressor = groupcompress.GroupCompressor()
        start = 0
        for key in sorted(key_to_text):
            compressor.compress(
                key, [key_to_text[key]], len(key_to_text[key]), None)
        locs = {key: (start, end) for key, (start, _, end, _)
                    in compressor.labels_deltas.items()}
        block = compressor.flush()
        raw_bytes = block.to_bytes()
        return locs, groupcompress.GroupCompressBlock.from_bytes(raw_bytes)

    def add_key_to_manager(self, key, locations, block, manager):
        start, end = locations[key]
        manager.add_factory(key, (), start, end)

    def make_block_and_full_manager(self, texts):
        locations, block = self.make_block(texts)
        manager = groupcompress._LazyGroupContentManager(block)
        for key in sorted(texts):
            self.add_key_to_manager(key, locations, block, manager)
        return block, manager

    def test_get_fulltexts(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b'key1',), locations, block, manager)
        self.add_key_to_manager((b'key2',), locations, block, manager)
        result_order = []
        for record in manager.get_record_stream():
            result_order.append(record.key)
            text = self._texts[record.key]
            self.assertEqual(text, record.get_bytes_as('fulltext'))
        self.assertEqual([(b'key1',), (b'key2',)], result_order)

        # If we build the manager in the opposite order, we should get them
        # back in the opposite order
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b'key2',), locations, block, manager)
        self.add_key_to_manager((b'key1',), locations, block, manager)
        result_order = []
        for record in manager.get_record_stream():
            result_order.append(record.key)
            text = self._texts[record.key]
            self.assertEqual(text, record.get_bytes_as('fulltext'))
        self.assertEqual([(b'key2',), (b'key1',)], result_order)

    def test__wire_bytes_no_keys(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        wire_bytes = manager._wire_bytes()
        block_length = len(block.to_bytes())
        # We should have triggered a strip, since we aren't using any content
        stripped_block = manager._block.to_bytes()
        self.assertTrue(block_length > len(stripped_block))
        empty_z_header = zlib.compress(b'')
        self.assertEqual(b'groupcompress-block\n'
                         b'8\n'  # len(compress(''))
                         b'0\n'  # len('')
                         b'%d\n'  # compressed block len
                         b'%s'  # zheader
                         b'%s'  # block
                         % (len(stripped_block), empty_z_header,
                            stripped_block),
                         wire_bytes)

    def test__wire_bytes(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b'key1',), locations, block, manager)
        self.add_key_to_manager((b'key4',), locations, block, manager)
        block_bytes = block.to_bytes()
        wire_bytes = manager._wire_bytes()
        (storage_kind, z_header_len, header_len,
         block_len, rest) = wire_bytes.split(b'\n', 4)
        z_header_len = int(z_header_len)
        header_len = int(header_len)
        block_len = int(block_len)
        self.assertEqual(b'groupcompress-block', storage_kind)
        self.assertEqual(34, z_header_len)
        self.assertEqual(26, header_len)
        self.assertEqual(len(block_bytes), block_len)
        z_header = rest[:z_header_len]
        header = zlib.decompress(z_header)
        self.assertEqual(header_len, len(header))
        entry1 = locations[(b'key1',)]
        entry4 = locations[(b'key4',)]
        self.assertEqualDiff(b'key1\n'
                             b'\n'  # no parents
                             b'%d\n'  # start offset
                             b'%d\n'  # end offset
                             b'key4\n'
                             b'\n'
                             b'%d\n'
                             b'%d\n'
                             % (entry1[0], entry1[1],
                                entry4[0], entry4[1]),
                             header)
        z_block = rest[z_header_len:]
        self.assertEqual(block_bytes, z_block)

    def test_from_bytes(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b'key1',), locations, block, manager)
        self.add_key_to_manager((b'key4',), locations, block, manager)
        wire_bytes = manager._wire_bytes()
        self.assertStartsWith(wire_bytes, b'groupcompress-block\n')
        manager = groupcompress._LazyGroupContentManager.from_bytes(wire_bytes)
        self.assertIsInstance(manager, groupcompress._LazyGroupContentManager)
        self.assertEqual(2, len(manager._factories))
        self.assertEqual(block._z_content, manager._block._z_content)
        result_order = []
        for record in manager.get_record_stream():
            result_order.append(record.key)
            text = self._texts[record.key]
            self.assertEqual(text, record.get_bytes_as('fulltext'))
        self.assertEqual([(b'key1',), (b'key4',)], result_order)

    def test__check_rebuild_no_changes(self):
        block, manager = self.make_block_and_full_manager(self._texts)
        manager._check_rebuild_block()
        self.assertIs(block, manager._block)

    def test__check_rebuild_only_one(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        # Request just the first key, which should trigger a 'strip' action
        self.add_key_to_manager((b'key1',), locations, block, manager)
        manager._check_rebuild_block()
        self.assertIsNot(block, manager._block)
        self.assertTrue(block._content_length > manager._block._content_length)
        # We should be able to still get the content out of this block, though
        # it should only have 1 entry
        for record in manager.get_record_stream():
            self.assertEqual((b'key1',), record.key)
            self.assertEqual(self._texts[record.key],
                             record.get_bytes_as('fulltext'))

    def test__check_rebuild_middle(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        # Request a small key in the middle should trigger a 'rebuild'
        self.add_key_to_manager((b'key4',), locations, block, manager)
        manager._check_rebuild_block()
        self.assertIsNot(block, manager._block)
        self.assertTrue(block._content_length > manager._block._content_length)
        for record in manager.get_record_stream():
            self.assertEqual((b'key4',), record.key)
            self.assertEqual(self._texts[record.key],
                             record.get_bytes_as('fulltext'))

    def test_manager_default_compressor_settings(self):
        locations, old_block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(old_block)
        gcvf = groupcompress.GroupCompressVersionedFiles
        # It doesn't greedily evaluate _max_bytes_to_index
        self.assertIs(None, manager._compressor_settings)
        self.assertEqual(gcvf._DEFAULT_COMPRESSOR_SETTINGS,
                         manager._get_compressor_settings())

    def test_manager_custom_compressor_settings(self):
        locations, old_block = self.make_block(self._texts)
        called = []

        def compressor_settings():
            called.append('called')
            return (10,)
        manager = groupcompress._LazyGroupContentManager(old_block,
                                                         get_compressor_settings=compressor_settings)
        gcvf = groupcompress.GroupCompressVersionedFiles
        # It doesn't greedily evaluate compressor_settings
        self.assertIs(None, manager._compressor_settings)
        self.assertEqual((10,), manager._get_compressor_settings())
        self.assertEqual((10,), manager._get_compressor_settings())
        self.assertEqual((10,), manager._compressor_settings)
        # Only called 1 time
        self.assertEqual(['called'], called)

    def test__rebuild_handles_compressor_settings(self):
        if not isinstance(groupcompress.GroupCompressor,
                          groupcompress.PyrexGroupCompressor):
            raise tests.TestNotApplicable('pure-python compressor'
                                          ' does not handle compressor_settings')
        locations, old_block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(old_block,
                                                         get_compressor_settings=lambda: dict(max_bytes_to_index=32))
        gc = manager._make_group_compressor()
        self.assertEqual(32, gc._delta_index._max_bytes_to_index)
        self.add_key_to_manager((b'key3',), locations, old_block, manager)
        self.add_key_to_manager((b'key4',), locations, old_block, manager)
        action, last_byte, total_bytes = manager._check_rebuild_action()
        self.assertEqual('rebuild', action)
        manager._rebuild_block()
        new_block = manager._block
        self.assertIsNot(old_block, new_block)
        # Because of the new max_bytes_to_index, we do a poor job of
        # rebuilding. This is a side-effect of the change, but at least it does
        # show the setting had an effect.
        self.assertTrue(old_block._content_length < new_block._content_length)

    def test_check_is_well_utilized_all_keys(self):
        block, manager = self.make_block_and_full_manager(self._texts)
        self.assertFalse(manager.check_is_well_utilized())
        # Though we can fake it by changing the recommended minimum size
        manager._full_enough_block_size = block._content_length
        self.assertTrue(manager.check_is_well_utilized())
        # Setting it just above causes it to fail
        manager._full_enough_block_size = block._content_length + 1
        self.assertFalse(manager.check_is_well_utilized())
        # Setting the mixed-block size doesn't do anything, because the content
        # is considered to not be 'mixed'
        manager._full_enough_mixed_block_size = block._content_length
        self.assertFalse(manager.check_is_well_utilized())

    def test_check_is_well_utilized_mixed_keys(self):
        texts = {}
        f1k1 = (b'f1', b'k1')
        f1k2 = (b'f1', b'k2')
        f2k1 = (b'f2', b'k1')
        f2k2 = (b'f2', b'k2')
        texts[f1k1] = self._texts[(b'key1',)]
        texts[f1k2] = self._texts[(b'key2',)]
        texts[f2k1] = self._texts[(b'key3',)]
        texts[f2k2] = self._texts[(b'key4',)]
        block, manager = self.make_block_and_full_manager(texts)
        self.assertFalse(manager.check_is_well_utilized())
        manager._full_enough_block_size = block._content_length
        self.assertTrue(manager.check_is_well_utilized())
        manager._full_enough_block_size = block._content_length + 1
        self.assertFalse(manager.check_is_well_utilized())
        manager._full_enough_mixed_block_size = block._content_length
        self.assertTrue(manager.check_is_well_utilized())

    def test_check_is_well_utilized_partial_use(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        manager._full_enough_block_size = block._content_length
        self.add_key_to_manager((b'key1',), locations, block, manager)
        self.add_key_to_manager((b'key2',), locations, block, manager)
        # Just using the content from key1 and 2 is not enough to be considered
        # 'complete'
        self.assertFalse(manager.check_is_well_utilized())
        # However if we add key3, then we have enough, as we only require 75%
        # consumption
        self.add_key_to_manager((b'key4',), locations, block, manager)
        self.assertTrue(manager.check_is_well_utilized())


class Test_GCBuildDetails(tests.TestCase):

    def test_acts_like_tuple(self):
        # _GCBuildDetails inlines some of the data that used to be spread out
        # across a bunch of tuples
        bd = groupcompress._GCBuildDetails((('parent1',), ('parent2',)),
                                           ('INDEX', 10, 20, 0, 5))
        self.assertEqual(4, len(bd))
        self.assertEqual(('INDEX', 10, 20, 0, 5), bd[0])
        self.assertEqual(None, bd[1])  # Compression Parent is always None
        self.assertEqual((('parent1',), ('parent2',)), bd[2])
        self.assertEqual(('group', None), bd[3])  # Record details

    def test__repr__(self):
        bd = groupcompress._GCBuildDetails((('parent1',), ('parent2',)),
                                           ('INDEX', 10, 20, 0, 5))
        self.assertEqual("_GCBuildDetails(('INDEX', 10, 20, 0, 5),"
                         " (('parent1',), ('parent2',)))",
                         repr(bd))



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for get_rev_id_for_revno on a repository with external references."""

from breezy import errors
from breezy.tests.per_repository_reference import \
    TestCaseWithExternalReferenceRepository


class TestGetRevIdForRevno(TestCaseWithExternalReferenceRepository):

    def test_uses_fallback(self):
        tree = self.make_branch_and_tree('base')
        base = tree.branch.repository
        revid = tree.commit('one')
        revid2 = tree.commit('two')
        spare_tree = tree.controldir.sprout('spare').open_workingtree()
        revid3 = spare_tree.commit('three')
        branch = spare_tree.branch.create_clone_on_transport(
            self.get_transport('referring'),
            stacked_on=tree.branch.base)
        repo = branch.repository
        # Sanity check: now repo has 'revid3', and base has 'revid' + 'revid2'
        self.assertEqual({revid3},
                         set(repo.controldir.open_repository().all_revision_ids()))
        self.assertEqual({revid2, revid},
                         set(base.controldir.open_repository().all_revision_ids()))
        # get_rev_id_for_revno will find revno 1 == 'revid', even though
        # that revision can only be found in the fallback.
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual(
            (True, revid), repo.get_rev_id_for_revno(1, (3, revid3)))



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import errors, osutils
from . import inventory, xml6
from .xml_serializer import (encode_and_escape, get_utf8_or_ascii,
                             unpack_inventory_entry)


class Serializer_v5(xml6.Serializer_v6):
    """Version 5 serializer

    Packs objects into XML and vice versa.
    """
    format_num = b'5'
    root_id = inventory.ROOT_ID

    def _unpack_inventory(self, elt, revision_id, entry_cache=None,
                          return_from_cache=False):
        """Construct from XML Element
        """
        root_id = elt.get('file_id') or inventory.ROOT_ID
        root_id = get_utf8_or_ascii(root_id)

        format = elt.get('format')
        if format is not None:
            if format != '5':
                raise errors.BzrError("invalid format version %r on inventory"
                                      % format)
        data_revision_id = elt.get('revision_id')
        if data_revision_id is not None:
            revision_id = data_revision_id.encode('utf-8')
        inv = inventory.Inventory(root_id, revision_id=revision_id)
        # Optimizations tested
        #   baseline w/entry cache  2.85s
        #   using inv._byid         2.55s
        #   avoiding attributes     2.46s
        #   adding assertions       2.50s
        #   last_parent cache       2.52s (worse, removed)
        byid = inv._byid
        for e in elt:
            ie = unpack_inventory_entry(e, entry_cache=entry_cache,
                                        return_from_cache=return_from_cache)
            parent_id = ie.parent_id
            if parent_id is None:
                ie.parent_id = parent_id = root_id
            try:
                parent = byid[parent_id]
            except KeyError:
                raise errors.BzrError("parent_id {%s} not in inventory"
                                      % (parent_id,))
            if ie.file_id in byid:
                raise inventory.DuplicateFileId(ie.file_id, byid[ie.file_id])
            if ie.name in parent.children:
                raise errors.BzrError(
                    "{} is already versioned".format(
                        osutils.pathjoin(
                            inv.id2path(parent_id), ie.name).encode('utf-8')))
            parent.children[ie.name] = ie
            byid[ie.file_id] = ie
        if revision_id is not None:
            inv.root.revision = revision_id
        self._check_cache_size(len(inv), entry_cache)
        return inv

    def _check_revisions(self, inv):
        """Extension point for subclasses to check during serialisation.

        In this version, no checking is done.

        :param inv: An inventory about to be serialised, to be checked.
        :raises: AssertionError if an error has occurred.
        """

    def _append_inventory_root(self, append, inv):
        """Append the inventory root to output."""
        if inv.root.file_id not in (None, inventory.ROOT_ID):
            fileid = b''.join([b' file_id="', encode_and_escape(inv.root.file_id), b'"'])
        else:
            fileid = b""
        if inv.revision_id is not None:
            revid = b''.join([b' revision_id="', encode_and_escape(inv.revision_id), b'"'])
        else:
            revid = b""
        append(b'<inventory%s format="5"%s>\n' % (fileid, revid))


serializer_v5 = Serializer_v5()
